import jax
import chex
from .restarter import RestartWrapper
from .termination import spread_criterion
from flax import struct


@struct.dataclass
class RestartParams:
    min_num_gens: int = 50
    min_fitness_spread: float = 0.1
    copy_mean: bool = False


class Simple_Restarter(RestartWrapper):
    def __init__(
        self,
        base_strategy,
        stop_criteria=[spread_criterion],
    ):
        """Simple Restart Strategy - Only reinitialize the state."""
        super().__init__(base_strategy, stop_criteria)

    @property
    def restart_params(self) -> RestartParams:
        """Return default parameters for strategy restarting."""
        return RestartParams()

    def restart_strategy(
        self,
        rng: chex.PRNGKey,
        state: chex.ArrayTree,
        params: chex.ArrayTree,
    ) -> chex.ArrayTree:
        """Simple restart by state initialization."""
        new_state = self.base_strategy.initialize(rng, params.strategy_params)
        new_state = new_state.replace(
            mean=jax.lax.select(
                params.restart_params.copy_mean,
                state.strategy_state.mean,
                new_state.mean,
            )
        )
        return new_state



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    temp: float
    replace_rng: float
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    temp_init: float = 1.0
    temp_limit: float = 0.1
    temp_decay: float = 0.999
    boltzmann_const: float = 5.0
    sigma_init: float = 0.05
    sigma_limit: float = 0.001
    sigma_decay: float = 0.999
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class SimAnneal(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        sigma_init: float = 0.03,
        sigma_decay: float = 1.0,
        sigma_limit: float = 0.01,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Simulated Annealing (Rasdi Rere et al., 2015)
        Reference: https://www.sciencedirect.com/science/article/pii/S1877050915035759
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.strategy_name = "SimAnneal"

        # Set core kwargs es_params (lrate/sigma schedules)
        self.sigma_init = sigma_init
        self.sigma_decay = sigma_decay
        self.sigma_limit = sigma_limit

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(
            sigma_init=self.sigma_init,
            sigma_decay=self.sigma_decay,
            sigma_limit=self.sigma_limit,
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        rng_init, rng_rep = jax.random.split(rng)
        initialization = jax.random.uniform(
            rng_init,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init,
            temp=params.temp_init,
            replace_rng=jax.random.uniform(rng_rep, ()),
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        rng_noise, rng_rep = jax.random.split(rng)
        # Sampling of N(0, 1) noise
        z = jax.random.normal(
            rng_noise,
            (self.popsize, self.num_dims),
        )
        x = state.mean + state.sigma * z
        return x, state.replace(replace_rng=jax.random.uniform(rng_rep, ()))

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        best_in_gen = jnp.argmin(fitness)
        gen_fitness, gen_member = fitness[best_in_gen], x[best_in_gen]
        improve_diff = state.best_fitness - gen_fitness
        improved = improve_diff > 0

        # Calculate temperature replacement constant (replace by best in gen)
        metropolis = jnp.exp(
            improve_diff / (state.temp * params.boltzmann_const)
        )

        # Replace mean either if improvement or random metropolis acceptance
        rand_replace = jnp.logical_or(improved, state.replace_rng > metropolis)
        # Note: We replace by best member in generation (not completely random)
        mean = jax.lax.select(rand_replace, gen_member, state.mean)

        # Update permutation standard deviation
        sigma = jax.lax.select(
            state.sigma > params.sigma_limit,
            state.sigma * params.sigma_decay,
            state.sigma,
        )

        temp = jax.lax.select(
            state.temp > params.temp_limit,
            state.temp * params.temp_decay,
            state.temp,
        )
        return state.replace(mean=mean, sigma=sigma, temp=temp)



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from functools import partial
from flax import struct
from ..strategy import Strategy
from ..core import GradientOptimizer, OptState, OptParams, exp_decay
from ..utils import get_best_fitness_member


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    opt_state: OptState
    grad_subspace: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    opt_params: OptParams
    sigma_init: float = 0.03
    sigma_decay: float = 1.0
    sigma_limit: float = 0.01
    alpha: float = 0.5
    beta: float = 1.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class GuidedES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        subspace_dims: int = 1,  # k param in example notebook
        opt_name: str = "sgd",
        lrate_init: float = 0.05,
        lrate_decay: float = 1.0,
        lrate_limit: float = 0.001,
        sigma_init: float = 0.03,
        sigma_decay: float = 1.0,
        sigma_limit: float = 0.01,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float],
    ):
        """Guided ES (Maheswaranathan et al., 2018)
        Reference: https://arxiv.org/abs/1806.10230
        Note that there are a couple of JAX-based adaptations:
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs,
        )
        assert not self.popsize & 1, "Population size must be even"
        assert opt_name in ["sgd", "adam", "rmsprop", "clipup", "adan"]
        assert (
            subspace_dims <= self.num_dims
        ), "Subspace has to be smaller than optimization dims."
        self.optimizer = GradientOptimizer[opt_name](self.num_dims)
        self.subspace_dims = min(subspace_dims, self.num_dims)
        if self.subspace_dims < subspace_dims:
            print(
                "Subspace has to be smaller than optimization dims. Set to"
                f" {self.subspace_dims} instead of {subspace_dims}."
            )
        self.strategy_name = "GuidedES"

        # Set core kwargs es_params (lrate/sigma schedules)
        self.lrate_init = lrate_init
        self.lrate_decay = lrate_decay
        self.lrate_limit = lrate_limit
        self.sigma_init = sigma_init
        self.sigma_decay = sigma_decay
        self.sigma_limit = sigma_limit

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(opt_params=self.optimizer.default_params)

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        rng_init, rng_sub = jax.random.split(rng)
        initialization = jax.random.uniform(
            rng_init,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )

        grad_subspace = jax.random.normal(
            rng_sub, (self.subspace_dims, self.num_dims)
        )

        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init,
            opt_state=self.optimizer.initialize(params.opt_params),
            grad_subspace=grad_subspace,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        a = state.sigma * jnp.sqrt(params.alpha / self.num_dims)
        c = state.sigma * jnp.sqrt((1.0 - params.alpha) / self.subspace_dims)
        key_full, key_sub = jax.random.split(rng, 2)
        eps_full = jax.random.normal(
            key_full, shape=(self.num_dims, int(self.popsize / 2))
        )
        eps_subspace = jax.random.normal(
            key_sub, shape=(self.subspace_dims, int(self.popsize / 2))
        )
        Q, _ = jnp.linalg.qr(state.grad_subspace)
        # Antithetic sampling of noise
        z_plus = a * eps_full + c * jnp.dot(Q, eps_subspace)
        z_plus = jnp.swapaxes(z_plus, 0, 1)
        z = jnp.concatenate([z_plus, -1.0 * z_plus])
        x = state.mean + z
        return x, state

    @partial(jax.jit, static_argnums=(0,))
    def tell(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: Optional[EvoParams] = None,
        gradient: Optional[chex.Array] = None,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Use default hyperparameters if no other settings provided
        if params is None:
            params = self.default_params

        # Flatten params if using param reshaper for ES update
        if self.use_param_reshaper:
            x = self.param_reshaper.flatten(x)

        # Perform fitness reshaping inside of strategy tell call (if desired)
        fitness_re = self.fitness_shaper.apply(x, fitness)

        # Reconstruct noise from last mean/std estimates
        noise = (x - state.mean) / state.sigma
        noise_1 = noise[: int(self.popsize / 2)]
        fit_1 = fitness_re[: int(self.popsize / 2)]
        fit_2 = fitness_re[int(self.popsize / 2) :]
        fit_diff = fit_1 - fit_2
        fit_diff_noise = jnp.dot(noise_1.T, fit_diff)
        theta_grad = (params.beta / self.popsize) * fit_diff_noise

        # Add grad FIFO-style to subspace archive (only if provided else FD)
        grad_subspace = jnp.zeros((self.subspace_dims, self.num_dims))
        grad_subspace = grad_subspace.at[:-1, :].set(state.grad_subspace[1:, :])
        if gradient is not None:
            grad_subspace = grad_subspace.at[-1, :].set(gradient)
        else:
            grad_subspace = grad_subspace.at[-1, :].set(theta_grad)
        state = state.replace(grad_subspace=grad_subspace)

        # Grad update using optimizer instance - decay lrate if desired
        mean, opt_state = self.optimizer.step(
            state.mean, theta_grad, state.opt_state, params.opt_params
        )
        opt_state = self.optimizer.update(opt_state, params.opt_params)

        # Update lrate and standard deviation based on min and decay
        sigma = exp_decay(state.sigma, params.sigma_decay, params.sigma_limit)
        state = state.replace(mean=mean, sigma=sigma, opt_state=opt_state)

        # Check if there is a new best member & update trackers
        best_member, best_fitness = get_best_fitness_member(x, fitness, state)
        return state.replace(
            best_member=best_member,
            best_fitness=best_fitness,
            gen_counter=state.gen_counter + 1,
        )



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from ..core import GradientOptimizer, OptState, OptParams, exp_decay


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: chex.Array
    opt_state: OptState
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    opt_params: OptParams
    sigma_init: float = 0.03
    sigma_decay: float = 0.999
    sigma_limit: float = 0.01
    sigma_lrate: float = 0.2  # Learning rate for std
    sigma_max_change: float = 0.2  # Clip adaptive sigma to 20%
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class ESMC(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        opt_name: str = "adam",
        lrate_init: float = 0.05,
        lrate_decay: float = 1.0,
        lrate_limit: float = 0.001,
        sigma_init: float = 0.03,
        sigma_decay: float = 1.0,
        sigma_limit: float = 0.01,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """ESMC (Merchant et al., 2021)
        Reference: https://proceedings.mlr.press/v139/merchant21a.html
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        assert self.popsize & 1, "Population size must be odd"
        assert opt_name in ["sgd", "adam", "rmsprop", "clipup", "adan"]
        self.optimizer = GradientOptimizer[opt_name](self.num_dims)
        self.strategy_name = "ESMC"

        # Set core kwargs es_params (lrate/sigma schedules)
        self.lrate_init = lrate_init
        self.lrate_decay = lrate_decay
        self.lrate_limit = lrate_limit
        self.sigma_init = sigma_init
        self.sigma_decay = sigma_decay
        self.sigma_limit = sigma_limit

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        opt_params = self.optimizer.default_params.replace(
            lrate_init=self.lrate_init,
            lrate_decay=self.lrate_decay,
            lrate_limit=self.lrate_limit,
        )
        return EvoParams(
            opt_params=opt_params,
            sigma_init=self.sigma_init,
            sigma_decay=self.sigma_decay,
            sigma_limit=self.sigma_limit,
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            sigma=jnp.ones(self.num_dims) * params.sigma_init,
            opt_state=self.optimizer.initialize(params.opt_params),
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        # Antithetic sampling of noise
        z_plus = jax.random.normal(
            rng,
            (int(self.popsize / 2), self.num_dims),
        )
        z = jnp.concatenate(
            [jnp.zeros((1, self.num_dims)), z_plus, -1.0 * z_plus]
        )
        x = state.mean + z * state.sigma.reshape(1, self.num_dims)
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """Update both mean and dim.-wise isotropic Gaussian scale."""
        # Reconstruct noise from last mean/std estimates
        noise = (x - state.mean) / state.sigma
        bline_fitness = fitness[0]
        noise = noise[1:]
        fitness = fitness[1:]
        noise_1 = noise[: int((self.popsize - 1) / 2)]
        fit_1 = fitness[: int((self.popsize - 1) / 2)]
        fit_2 = fitness[int((self.popsize - 1) / 2) :]
        fit_diff = jnp.minimum(fit_1, bline_fitness) - jnp.minimum(
            fit_2, bline_fitness
        )
        fit_diff_noise = jnp.dot(noise_1.T, fit_diff)
        theta_grad = 1.0 / int((self.popsize - 1) / 2) * fit_diff_noise
        # Grad update using optimizer instance - decay lrate if desired
        mean, opt_state = self.optimizer.step(
            state.mean, theta_grad, state.opt_state, params.opt_params
        )
        opt_state = self.optimizer.update(opt_state, params.opt_params)
        sigma = state.sigma * params.sigma_decay
        sigma = jnp.maximum(sigma, params.sigma_limit)
        return state.replace(mean=mean, sigma=sigma, opt_state=opt_state)



# Import additional utilities - Logging, visualization
from .es_logger import ESLog

# Import general helper utilities
from .helpers import get_best_fitness_member

# 2D Fitness visualization tools
from .visualizer_2d import BBOBVisualizer


__all__ = [
    "get_best_fitness_member",
    "ESLog",
    "BBOBVisualizer",
]



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from .simple_ga import single_mate
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    archive: chex.Array
    fitness: chex.Array
    sigma: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    cross_over_rate: float = 0.0
    sigma_init: float = 0.07
    sigma_ratio: float = 0.15
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class MR15_GA(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.0,
        sigma_ratio: float = 0.15,
        sigma_init: float = 0.1,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """1/5 MR Genetic Algorithm (Rechenberg, 1987)
        Reference: https://link.springer.com/chapter/10.1007/978-3-642-81283-5_8
        """

        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.strategy_name = "MR15_GA"

        # Set core kwargs es_params
        self.sigma_ratio = sigma_ratio  # no. mutation that have to improve
        self.sigma_init = sigma_init

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(
            sigma_init=self.sigma_init, sigma_ratio=self.sigma_ratio
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the differential evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.elite_popsize, self.num_dims),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization.mean(axis=0),
            archive=initialization,
            fitness=jnp.zeros(self.elite_popsize) + jnp.finfo(jnp.float32).max,
            sigma=params.sigma_init,
            best_member=initialization.mean(axis=0),
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """
        `ask` for new proposed candidates to evaluate next.
        1. For each member of elite:
          - Sample two current elite members (a & b)
          - Cross over all dims of a with corresponding one from b
            if random number > co-rate
          - Additionally add noise on top of all elite parameters
        """
        rng, rng_eps, rng_idx_a, rng_idx_b = jax.random.split(rng, 4)
        rng_mate = jax.random.split(rng, self.popsize)
        epsilon = (
            jax.random.normal(rng_eps, (self.popsize, self.num_dims))
            * state.sigma
        )
        elite_ids = jnp.arange(self.elite_popsize)
        idx_a = jax.random.choice(rng_idx_a, elite_ids, (self.popsize,))
        idx_b = jax.random.choice(rng_idx_b, elite_ids, (self.popsize,))
        members_a = state.archive[idx_a]
        members_b = state.archive[idx_b]
        x = jax.vmap(single_mate, in_axes=(0, 0, 0, None))(
            rng_mate, members_a, members_b, params.cross_over_rate
        )
        x += epsilon
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """
        `tell` update to ES state.
        If fitness of y <= fitness of x -> replace in population.
        """
        # Combine current elite and recent generation info
        fitness = jnp.concatenate([fitness, state.fitness])
        solution = jnp.concatenate([x, state.archive])
        # Select top elite from total archive info
        idx = jnp.argsort(fitness)[0 : self.elite_popsize]
        fitness = fitness[idx]
        archive = solution[idx]
        # Update mutation sigma - double if more than 15% improved
        good_mutations_ratio = jnp.mean(fitness < state.best_fitness)
        increase_sigma = good_mutations_ratio > params.sigma_ratio
        sigma = jax.lax.select(
            increase_sigma, 2 * state.sigma, 0.5 * state.sigma
        )
        # Set mean to best member seen so far
        improved = fitness[0] < state.best_fitness
        best_mean = jax.lax.select(improved, archive[0], state.best_member)
        return state.replace(
            fitness=fitness, archive=archive, sigma=sigma, mean=best_mean
        )



from .mlp import MLP
from .cnn import CNN, All_CNN_C
from .lstm import LSTM


# Helper that returns model based on string name
NetworkMapper = {
    "MLP": MLP,
    "CNN": CNN,
    "All_CNN_C": All_CNN_C,
    "LSTM": LSTM,
}

__all__ = [
    "MLP",
    "CNN",
    "All_CNN_C",
    "LSTM",
    "NetworkMapper",
]



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from ..core import GradientOptimizer, OptState, OptParams, exp_decay


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: float
    opt_state: OptState
    grad_subspace: chex.Array
    alpha: float
    UUT: chex.Array
    UUT_ort: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    opt_params: OptParams
    sigma_init: float = 0.03
    sigma_decay: float = 1.0
    sigma_limit: float = 0.01
    grad_decay: float = 0.99
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class ASEBO(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        subspace_dims: int = 50,
        opt_name: str = "adam",
        lrate_init: float = 0.05,
        lrate_decay: float = 1.0,
        lrate_limit: float = 0.001,
        sigma_init: float = 0.03,
        sigma_decay: float = 1.0,
        sigma_limit: float = 0.01,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float],
    ):
        """ASEBO (Choromanski et al., 2019)
        Reference: https://arxiv.org/abs/1903.04268
        Note that there are a couple of JAX-based adaptations:
        1. We always sample a fixed population size per generation
        2. We keep a fixed archive of gradients to estimate the subspace
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs,
        )
        assert not self.popsize & 1, "Population size must be even"
        assert opt_name in ["sgd", "adam", "rmsprop", "clipup"]
        self.optimizer = GradientOptimizer[opt_name](self.num_dims)
        self.subspace_dims = min(subspace_dims, self.num_dims)
        if self.subspace_dims < subspace_dims:
            print(
                "Subspace has to be smaller than optimization dims. Set to"
                f" {self.subspace_dims} instead of {subspace_dims}."
            )
        self.strategy_name = "ASEBO"

        # Set core kwargs es_params (lrate/sigma schedules)
        self.lrate_init = lrate_init
        self.lrate_decay = lrate_decay
        self.lrate_limit = lrate_limit
        self.sigma_init = sigma_init
        self.sigma_decay = sigma_decay
        self.sigma_limit = sigma_limit

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        opt_params = self.optimizer.default_params.replace(
            lrate_init=self.lrate_init,
            lrate_decay=self.lrate_decay,
            lrate_limit=self.lrate_limit,
        )
        return EvoParams(
            opt_params=opt_params,
            sigma_init=self.sigma_init,
            sigma_decay=self.sigma_decay,
            sigma_limit=self.sigma_limit,
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )

        grad_subspace = jnp.zeros((self.subspace_dims, self.num_dims))

        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init,
            opt_state=self.optimizer.initialize(params.opt_params),
            grad_subspace=grad_subspace,
            alpha=1.0,
            UUT=jnp.zeros((self.num_dims, self.num_dims)),
            UUT_ort=jnp.zeros((self.num_dims, self.num_dims)),
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        # Antithetic sampling of noise
        X = state.grad_subspace
        X -= jnp.mean(X, axis=0)
        U, S, Vt = jnp.linalg.svd(X, full_matrices=False)

        def svd_flip(u, v):
            # columns of u, rows of v
            max_abs_cols = jnp.argmax(jnp.abs(u), axis=0)
            signs = jnp.sign(u[max_abs_cols, jnp.arange(u.shape[1])])
            u *= signs
            v *= signs[:, jnp.newaxis]
            return u, v

        U, Vt = svd_flip(U, Vt)
        U = Vt[: int(self.popsize / 2)]
        UUT = jnp.matmul(U.T, U)

        U_ort = Vt[int(self.popsize / 2) :]
        UUT_ort = jnp.matmul(U_ort.T, U_ort)

        subspace_ready = state.gen_counter > self.subspace_dims

        UUT = jax.lax.select(
            subspace_ready, UUT, jnp.zeros((self.num_dims, self.num_dims))
        )
        cov = (
            state.sigma * (state.alpha / self.num_dims) * jnp.eye(self.num_dims)
            + ((1 - state.alpha) / int(self.popsize / 2)) * UUT
        )
        chol = jnp.linalg.cholesky(cov)
        noise = jax.random.normal(rng, (self.num_dims, int(self.popsize / 2)))
        z_plus = jnp.swapaxes(chol @ noise, 0, 1)
        z_plus /= jnp.linalg.norm(z_plus, axis=-1)[:, jnp.newaxis]
        z = jnp.concatenate([z_plus, -1.0 * z_plus])
        x = state.mean + z
        return x, state.replace(UUT=UUT, UUT_ort=UUT_ort)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Reconstruct noise from last mean/std estimates
        noise = (x - state.mean) / state.sigma
        noise_1 = noise[: int(self.popsize / 2)]
        fit_1 = fitness[: int(self.popsize / 2)]
        fit_2 = fitness[int(self.popsize / 2) :]
        fit_diff_noise = jnp.dot(noise_1.T, fit_1 - fit_2)
        theta_grad = 1.0 / 2.0 * fit_diff_noise

        alpha = jnp.linalg.norm(
            jnp.dot(theta_grad, state.UUT_ort)
        ) / jnp.linalg.norm(jnp.dot(theta_grad, state.UUT))
        subspace_ready = state.gen_counter > self.subspace_dims
        alpha = jax.lax.select(subspace_ready, alpha, 1.0)

        # Add grad FIFO-style to subspace archive (only if provided else FD)
        grad_subspace = jnp.zeros((self.subspace_dims, self.num_dims))
        grad_subspace = grad_subspace.at[:-1, :].set(state.grad_subspace[1:, :])
        grad_subspace = grad_subspace.at[-1, :].set(theta_grad)
        state = state.replace(grad_subspace=grad_subspace)

        # Normalize gradients by norm / num_dims
        theta_grad /= jnp.linalg.norm(theta_grad) / self.num_dims + 1e-8

        # Grad update using optimizer instance - decay lrate if desired
        mean, opt_state = self.optimizer.step(
            state.mean, theta_grad, state.opt_state, params.opt_params
        )
        opt_state = self.optimizer.update(opt_state, params.opt_params)

        # Update lrate and standard deviation based on min and decay
        sigma = state.sigma * params.sigma_decay
        sigma = exp_decay(state.sigma, params.sigma_decay, params.sigma_limit)
        return state.replace(
            mean=mean, sigma=sigma, opt_state=opt_state, alpha=alpha
        )



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: chex.Array
    weights: chex.Array  # Weights for population members
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    c_sigma: float = 0.1  # Learning rate for population std
    c_m: float = 1.0  # Learning rate for population mean
    sigma_init: float = 1.0  # Standard deviation
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class SimpleES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.5,
        sigma_init: float = 1.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Simple Gaussian Evolution Strategy (Rechenberg, 1975)
        Reference: https://onlinelibrary.wiley.com/doi/abs/10.1002/fedr.19750860506
        Inspired by: https://github.com/hardmaru/estool/blob/master/es.py"""
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs
        )
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.strategy_name = "SimpleES"

        # Set core kwargs es_params
        self.sigma_init = sigma_init

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        # Only parents have positive weight - equal weighting!
        return EvoParams(sigma_init=self.sigma_init)

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        weights = jnp.zeros(self.popsize)
        weights = weights.at[: self.elite_popsize].set(1 / self.elite_popsize)

        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            sigma=jnp.repeat(params.sigma_init, self.num_dims),
            weights=weights,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        z = jax.random.normal(rng, (self.popsize, self.num_dims))  # ~ N(0, I)
        x = state.mean + state.sigma * z  # ~ N(m, σ^2 I)
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        # Sort new results, extract elite, store best performer
        concat_p_f = jnp.hstack([jnp.expand_dims(fitness, 1), x])
        sorted_solutions = concat_p_f[concat_p_f[:, 0].argsort()]
        # Update mean, isotropic/anisotropic paths, covariance, stepsize
        mean, y_k = update_mean(
            sorted_solutions, state.mean, params.c_m, state.weights
        )
        sigma = update_sigma(y_k, state.sigma, params.c_sigma, state.weights)
        return state.replace(mean=mean, sigma=sigma)


def update_mean(
    sorted_solutions: chex.Array,
    mean: chex.Array,
    c_m: float,
    weights: chex.Array,
) -> Tuple[chex.Array, chex.Array]:
    """Update mean of strategy."""
    x_k = sorted_solutions[:, 1:]  # ~ N(m, σ^2 C)
    y_k = x_k - mean
    y_w = jnp.sum(y_k.T * weights, axis=1)
    mean_new = mean + c_m * y_w
    return mean_new, y_k


def update_sigma(
    y_k: chex.Array, sigma: chex.Array, c_sigma: float, weights: chex.Array
) -> chex.Array:
    """Update stepsize sigma."""
    sigma_est = jnp.sqrt(jnp.sum((y_k.T ** 2 * weights), axis=1))
    sigma_new = (1 - c_sigma) * sigma + c_sigma * sigma_est
    return sigma_new



from typing import Optional, Tuple, Union
import chex
from flax import struct
import jax
import jax.numpy as jnp
import pkgutil
from ..learned_eo.les_tools import (
    AttentionWeights,
    EvoPathMLP,
    tanh_timestamp,
    EvolutionPath,
    FitnessFeatures,
    load_pkl_object,
)
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: chex.Array
    path_c: chex.Array
    path_sigma: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    net_params: chex.ArrayTree
    sigma_init: float = 0.1
    init_min: float = -5.0
    init_max: float = 5.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class LES(Strategy):
    """Population-Invariant Learned Evolution Strategy (Lange et al., 2023)."""

    # NOTE: This is an independent reimplementation which does not use the same
    # meta-trained checkpoint used to generate the results in the paper. It
    # has been independently meta-trained & tested on a handful of Brax tasks.
    # The results may therefore differ from the ones shown in the paper.
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        net_params: Optional[chex.ArrayTree] = None,
        net_ckpt_path: Optional[str] = None,
        sigma_init: float = 0.1,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float],
    ):
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs,
        )
        self.strategy_name = "LES"
        self.evopath = EvolutionPath(
            num_dims=self.num_dims, timescales=jnp.array([0.1, 0.5, 0.9])
        )
        self.weight_layer = AttentionWeights(8)
        self.lrate_layer = EvoPathMLP(8)
        self.fitness_features = FitnessFeatures(
            centered_rank=True, z_score=True
        )
        self.sigma_init = sigma_init

        # Set net params provided at instantiation
        if net_params is not None:
            self.les_net_params = net_params

        # Load network weights from checkpoint
        if net_ckpt_path is not None:
            self.les_net_params = load_pkl_object(net_ckpt_path)
            print(f"Loaded LES model from ckpt: {net_ckpt_path}")

        if net_params is None and net_ckpt_path is None:
            ckpt_fname = "2023_10_les_v2.pkl"
            data = pkgutil.get_data(__name__, f"ckpt/les/{ckpt_fname}")
            self.les_net_params = load_pkl_object(data, pkg_load=True)
            print(f"Loaded pretrained LES model from ckpt: {ckpt_fname}")

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(
            net_params=self.les_net_params, sigma_init=self.sigma_init
        )

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        init_mean = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        init_sigma = params.sigma_init * jnp.ones(self.num_dims)
        init_path_c = self.evopath.initialize()
        init_path_sigma = self.evopath.initialize()
        return EvoState(
            mean=init_mean,
            sigma=init_sigma,
            path_c=init_path_c,
            path_sigma=init_path_sigma,
            best_member=init_mean,
        )

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        noise = jax.random.normal(rng, (self.popsize, self.num_dims))
        x = state.mean + noise * state.sigma.reshape(1, self.num_dims)
        x = jnp.clip(x, params.clip_min, params.clip_max)
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        fit_re = self.fitness_features.apply(x, fitness, state.best_fitness)
        time_embed = tanh_timestamp(state.gen_counter)
        weights = self.weight_layer.apply(
            params.net_params["recomb_weights"], fit_re
        )
        weight_diff = (weights * (x - state.mean)).sum(axis=0)
        weight_noise = (weights * (x - state.mean) / state.sigma).sum(axis=0)
        path_c = self.evopath.update(state.path_c, weight_diff)
        path_sigma = self.evopath.update(state.path_sigma, weight_noise)
        lrates_mean, lrates_sigma = self.lrate_layer.apply(
            params.net_params["lrate_modulation"],
            path_c,
            path_sigma,
            time_embed,
        )
        weighted_mean = (weights * x).sum(axis=0)
        weighted_sigma = jnp.sqrt(
            (weights * (x - state.mean) ** 2).sum(axis=0) + 1e-10
        )
        mean_change = lrates_mean * (weighted_mean - state.mean)
        sigma_change = lrates_sigma * (weighted_sigma - state.sigma)
        mean = state.mean + mean_change
        sigma = state.sigma + sigma_change
        mean = jnp.clip(mean, params.clip_min, params.clip_max)
        sigma = jnp.clip(sigma, 0, params.clip_max)
        return state.replace(
            mean=mean,
            sigma=sigma,
            path_c=path_c,
            path_sigma=path_sigma,
        )



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Optional


class VisionFitness(object):
    def __init__(
        self,
        task_name: str = "MNIST",
        batch_size: int = 1024,
        test: bool = False,
        n_devices: Optional[int] = None,
    ):
        self.task_name = task_name
        self.batch_size = batch_size
        self.steps_per_member = 1
        self.test = test
        self.num_classes = 10
        self.action_shape = 10
        data = get_array_data(self.task_name, self.test)
        self.dataloader = BatchLoader(*data, batch_size=self.batch_size)
        if n_devices is None:
            self.n_devices = jax.local_device_count()
        else:
            self.n_devices = n_devices

    def set_apply_fn(self, network):
        """Set the network forward function."""
        self.network = network
        self.rollout_pop = jax.vmap(self.rollout_ffw, in_axes=(None, 0))
        # pmap over popmembers if > 1 device is available - otherwise pmap
        if self.n_devices > 1:
            self.rollout = self.rollout_pmap
            print(
                f"VisionFitness: {self.n_devices} devices detected. Please make"
                " sure that the ES population size divides evenly across the"
                " number of devices to pmap/parallelize over."
            )
        else:
            self.rollout = jax.jit(self.rollout_vmap)

    def rollout_vmap(
        self, rng_input: chex.PRNGKey, network_params: chex.ArrayTree
    ):
        """Vectorize rollout. Reshape output correctly."""
        loss, acc = self.rollout_pop(rng_input, network_params)
        loss_re = loss.reshape(-1, 1)
        acc_re = acc.reshape(-1, 1)
        return loss_re, acc_re

    def rollout_pmap(
        self, rng_input: chex.PRNGKey, network_params: chex.ArrayTree
    ):
        """Parallelize rollout across devices. Split keys/reshape correctly."""
        keys_pmap = jnp.tile(rng_input, (self.n_devices, 1))
        loss_dev, acc_dev = jax.pmap(self.rollout_pop)(
            keys_pmap, network_params
        )
        loss_re = loss_dev.reshape(-1, 1)
        acc_re = acc_dev.reshape(-1, 1)
        return loss_re, acc_re

    def rollout_ffw(
        self, rng_input: chex.PRNGKey, network_params: chex.ArrayTree
    ) -> chex.ArrayTree:
        """Evaluate a network on a supervised learning task."""
        rng_net, rng_sample = jax.random.split(rng_input)
        X, y = self.dataloader.sample(rng_sample)
        y_pred = self.network(network_params, X, rng_net)
        loss, acc = loss_and_acc(y_pred, y, self.num_classes)
        # Return negative loss to maximize!
        return -1 * loss, acc

    @property
    def input_shape(self) -> Tuple[int]:
        """Get the shape of the observation."""
        return (1,) + self.dataloader.data_shape


def loss_and_acc(
    y_pred: chex.Array, y_true: chex.Array, num_classes: int
) -> Tuple[chex.Array, chex.Array]:
    """Compute cross-entropy loss and accuracy."""
    acc = jnp.mean(jnp.argmax(y_pred, axis=-1) == y_true)
    labels = jax.nn.one_hot(y_true, num_classes)
    loss = -jnp.sum(labels * jax.nn.log_softmax(y_pred))
    loss /= labels.shape[0]
    return loss, acc


class BatchLoader:
    def __init__(
        self,
        X: chex.Array,
        y: chex.Array,
        batch_size: int,
    ):
        self.X = X
        self.y = y
        self.data_shape = self.X.shape[1:]
        self.num_train_samples = X.shape[0]
        self.batch_size = batch_size

    def sample(self, key: chex.PRNGKey) -> Tuple[chex.Array, chex.Array]:
        """Sample a single batch of X, y data."""
        sample_idx = jax.random.choice(
            key,
            jnp.arange(self.num_train_samples),
            (self.batch_size,),
            replace=False,
        )
        return (
            jnp.take(self.X, sample_idx, axis=0),
            jnp.take(self.y, sample_idx, axis=0),
        )


def get_mnist_loaders(test: bool = False):
    try:
        import torch
        from torchvision import datasets, transforms
    except ModuleNotFoundError as err:
        raise ModuleNotFoundError(
            f"{err}. You need to install `torch` and `torchvision`"
            "to use the `VisionFitness` module."
        )

    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,)),
            transforms.Lambda(lambda x: x.permute(1, 2, 0)),
        ]
    )
    bs = 10000 if test else 60000
    loader = torch.utils.data.DataLoader(
        datasets.MNIST(
            "~/data", download=True, train=not test, transform=transform
        ),
        batch_size=bs,
        shuffle=False,
    )
    return loader


def get_fashion_loaders(test: bool = False):
    try:
        import torch
        from torchvision import datasets, transforms
    except ModuleNotFoundError as err:
        raise ModuleNotFoundError(
            f"{err}. You need to install `torch` and `torchvision`"
            "to use the `VisionFitness` module."
        )

    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,)),
            transforms.Lambda(lambda x: x.permute(1, 2, 0)),
        ]
    )

    bs = 10000 if test else 60000
    loader = torch.utils.data.DataLoader(
        datasets.FashionMNIST(
            "~/data", download=True, train=not test, transform=transform
        ),
        batch_size=bs,
        shuffle=False,
    )
    return loader


def get_cifar_loaders(test: bool = False):
    """Get PyTorch Data Loaders for CIFAR-10."""
    try:
        import torch
        from torchvision import datasets, transforms
    except ModuleNotFoundError as err:
        raise ModuleNotFoundError(
            f"{err}. You need to install `torch` and `torchvision`"
            "to use the `VisionFitness` module."
        )

    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(
                (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
            ),
            transforms.Lambda(lambda x: x.permute(1, 2, 0)),
        ]
    )

    bs = 10000 if test else 50000
    loader = torch.utils.data.DataLoader(
        datasets.CIFAR10(
            root="~/data", train=not test, download=True, transform=transform
        ),
        batch_size=bs,
        shuffle=False,
    )
    return loader


def normalize(data_tensor):
    """re-scale image values to [-1, 1]"""
    return (data_tensor / 255.0) * 2.0 - 1.0


def get_svhn_loaders(test: bool = False):
    """Get PyTorch Data Loaders for SVHN."""
    try:
        import torch
        from torchvision import datasets, transforms
    except ModuleNotFoundError as err:
        raise ModuleNotFoundError(
            f"{err}. You need to install `torch` and `torchvision`"
            "to use the `VisionFitness` module."
        )

    transform = [
        transforms.ToTensor(),
        transforms.Lambda(lambda x: normalize(x)),
        transforms.Lambda(lambda x: x.permute(1, 2, 0)),
    ]

    bs = 26032 if test else 73257
    loader = torch.utils.data.DataLoader(
        datasets.SVHN(
            root="~/data", train=not test, download=True, transform=transform
        ),
        batch_size=bs,
        shuffle=False,
    )
    return loader


def get_array_data(task_name: str = "MNIST", test: bool = False):
    """Get raw data arrays to subsample from."""
    if task_name == "MNIST":
        loader = get_mnist_loaders(test)
    elif task_name == "FashionMNIST":
        loader = get_fashion_loaders(test)
    elif task_name == "CIFAR10":
        loader = get_cifar_loaders(test)
    elif task_name == "SVHN":
        loader = get_svhn_loaders(test)
    else:
        raise ValueError("Dataset is not supported.")
    for _, (data, target) in enumerate(loader):
        break
    return jnp.array(data), jnp.array(target)



from .strategy import Strategy, EvoState, EvoParams
from .strategies import (
    SimpleGA,
    SimpleES,
    CMA_ES,
    DE,
    PSO,
    OpenES,
    PGPE,
    PBT,
    PersistentES,
    ARS,
    Sep_CMA_ES,
    BIPOP_CMA_ES,
    IPOP_CMA_ES,
    Full_iAMaLGaM,
    Indep_iAMaLGaM,
    MA_ES,
    LM_MA_ES,
    RmES,
    GLD,
    SimAnneal,
    SNES,
    xNES,
    ESMC,
    DES,
    SAMR_GA,
    GESMR_GA,
    GuidedES,
    ASEBO,
    CR_FM_NES,
    MR15_GA,
    RandomSearch,
    LES,
    LGA,
    NoiseReuseES,
    HillClimber,
    EvoTF_ES,
)
from .core import FitnessShaper, ParameterReshaper
from .utils import ESLog
from .networks import NetworkMapper
from .problems import ProblemMapper


Strategies = {
    "SimpleGA": SimpleGA,
    "SimpleES": SimpleES,
    "CMA_ES": CMA_ES,
    "DE": DE,
    "PSO": PSO,
    "OpenES": OpenES,
    "PGPE": PGPE,
    "PBT": PBT,
    "PersistentES": PersistentES,
    "ARS": ARS,
    "Sep_CMA_ES": Sep_CMA_ES,
    "BIPOP_CMA_ES": BIPOP_CMA_ES,
    "IPOP_CMA_ES": IPOP_CMA_ES,
    "Full_iAMaLGaM": Full_iAMaLGaM,
    "Indep_iAMaLGaM": Indep_iAMaLGaM,
    "MA_ES": MA_ES,
    "LM_MA_ES": LM_MA_ES,
    "RmES": RmES,
    "GLD": GLD,
    "SimAnneal": SimAnneal,
    "SNES": SNES,
    "xNES": xNES,
    "ESMC": ESMC,
    "DES": DES,
    "SAMR_GA": SAMR_GA,
    "GESMR_GA": GESMR_GA,
    "GuidedES": GuidedES,
    "ASEBO": ASEBO,
    "CR_FM_NES": CR_FM_NES,
    "MR15_GA": MR15_GA,
    "RandomSearch": RandomSearch,
    "LES": LES,
    "LGA": LGA,
    "NoiseReuseES": NoiseReuseES,
    "HillClimber": HillClimber,
    "EvoTF_ES": EvoTF_ES,
}

__all__ = [
    "Strategies",
    "EvoState",
    "EvoParams",
    "FitnessShaper",
    "ParameterReshaper",
    "ESLog",
    "NetworkMapper",
    "ProblemMapper",
    "Strategy",
    "SimpleGA",
    "SimpleES",
    "CMA_ES",
    "DE",
    "PSO",
    "OpenES",
    "PGPE",
    "PBT",
    "PersistentES",
    "ARS",
    "Sep_CMA_ES",
    "BIPOP_CMA_ES",
    "IPOP_CMA_ES",
    "Full_iAMaLGaM",
    "Indep_iAMaLGaM",
    "MA_ES",
    "LM_MA_ES",
    "RmES",
    "GLD",
    "SimAnneal",
    "SNES",
    "xNES",
    "ESMC",
    "DES",
    "SAMR_GA",
    "GESMR_GA",
    "GuidedES",
    "ASEBO",
    "CR_FM_NES",
    "MR15_GA",
    "RandomSearch",
    "LES",
    "LGA",
    "NoiseReuseES",
    "HillClimber",
    "EvoTF_ES",
]



import jax
import jax.numpy as jnp
from evosax.networks import LSTM, MLP, CNN
from evosax import ParameterReshaper


def test_flat_vector():
    rng = jax.random.PRNGKey(0)
    vec_params = jax.random.normal(rng, (2,))
    reshaper = ParameterReshaper(vec_params)
    assert reshaper.total_params == 2

    # Test population batch matrix reshaping
    test_params = jnp.zeros((100, 2))
    out = reshaper.reshape(test_params)
    assert out.shape == (100, 2)


# def test_reshape_lstm():
#     rng = jax.random.PRNGKey(1)
#     network = LSTM(
#         num_hidden_units=10,
#         num_output_units=1,
#         output_activation="identity",
#     )
#     pholder = jnp.zeros((10,))
#     carry_init = network.initialize_carry()
#     net_params = network.init(
#         rng,
#         x=pholder,
#         carry=carry_init,
#         rng=rng,
#     )

#     reshaper = ParameterReshaper(net_params)
#     # net_params["params"]["LSTMCell_0"].keys()
#     assert reshaper.total_params == 851

#     # Test population batch matrix reshaping
#     test_params = jnp.zeros((100, 851))
#     out = reshaper.reshape(test_params)
#     assert out["params"]["LSTMCell_0"]["hf"]["kernel"].shape == (100, 10, 10)

#     # Test single network vector reshaping
#     test_single = jnp.zeros(851)
#     out = reshaper.reshape_single(test_single)
#     assert out["params"]["LSTMCell_0"]["hf"]["kernel"].shape == (10, 10)


def test_reshape_mlp():
    rng = jax.random.PRNGKey(1)
    network = MLP(
        num_hidden_units=64,
        num_hidden_layers=2,
        num_output_units=1,
        hidden_activation="relu",
        output_activation="identity",
    )
    pholder = jnp.zeros((10,))
    net_params = network.init(
        rng,
        x=pholder,
        rng=rng,
    )

    reshaper = ParameterReshaper(net_params)
    assert reshaper.total_params == (10 * 64 + 64 + 64 * 64 + 64 + 64 + 1)

    # Test population batch matrix reshaping
    test_params = jnp.zeros((100, 4929))
    out = reshaper.reshape(test_params)
    assert out["params"]["Dense_0"]["kernel"].shape == (100, 10, 64)

    # Test single network vector reshaping
    test_single = jnp.zeros(4929)
    out = reshaper.reshape_single(test_single)
    assert out["params"]["Dense_0"]["kernel"].shape == (10, 64)


def test_reshape_cnn():
    rng = jax.random.PRNGKey(1)
    network = CNN(
        depth_1=1,
        depth_2=1,
        features_1=16,
        features_2=8,
        kernel_1=3,
        kernel_2=5,
        strides_1=1,
        strides_2=1,
        num_linear_layers=1,
        num_hidden_units=16,
        num_output_units=10,
    )
    pholder = jnp.zeros((1, 28, 28, 1))
    net_params = network.init(
        rng,
        x=pholder,
        rng=rng,
    )

    reshaper = ParameterReshaper(net_params)
    assert reshaper.total_params == 9826

    # Test population batch matrix reshaping
    test_params = jnp.zeros((100, 9826))
    out = reshaper.reshape(test_params)
    assert out["params"]["Conv_0"]["kernel"].shape == (100, 3, 3, 1, 16)

    # Test single network vector reshaping
    test_single = jnp.zeros(9826)
    out = reshaper.reshape_single(test_single)
    assert out["params"]["Conv_0"]["kernel"].shape == (3, 3, 1, 16)



from typing import Optional, Tuple, Union
import chex
from flax import struct
import jax
import jax.numpy as jnp
import pkgutil
from ..learned_eo.les_tools import (
    FitnessFeatures,
    load_pkl_object,
)
from ..learned_eo.lga_tools import (
    SelectionAttention,
    SamplingAttention,
    MutationAttention,
    tanh_age,
)
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    rng: chex.PRNGKey
    mean: chex.Array
    archive_age: chex.Array  # Parents: 'Age' counter
    archive_x: chex.Array  # Parents: Solution vectors
    archive_f: chex.Array  # Parents: Fitness scores
    archive_sigma: chex.Array  # Parents: Mutation strengths
    sigma_C: chex.Array  # Children: Mutation strengths
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    net_params: chex.ArrayTree
    cross_over_rate: float = 0.0
    sigma_init: float = 1.0
    init_min: float = -5.0
    init_max: float = 5.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class LGA(Strategy):
    """Population-Invariant Learned Genetic Algorithm (Lange et al., 2023)."""

    # NOTE: This is an independent reimplementation which does not use the same
    # meta-trained checkpoint used to generate the results in the paper. It
    # has been independently meta-trained & tested on a handful of Brax tasks.
    # The results may therefore differ from the ones shown in the paper.
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 1.0,
        net_params: Optional[chex.ArrayTree] = None,
        net_ckpt_path: Optional[str] = None,
        sigma_init: float = 1.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float],
    ):
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs,
        )
        self.strategy_name = "LGA"
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.fitness_features = FitnessFeatures(centered_rank=True, z_score=True)
        self.sigma_init = sigma_init
        self.selection_layer = SelectionAttention(2, 16)
        self.sampling_layer = SamplingAttention(2, 16)
        self.mutation_layer = MutationAttention(2, 16)

        # Set net params provided at instantiation
        if net_params is not None:
            self.lga_net_params = net_params

        # Load network weights from checkpoint
        if net_ckpt_path is not None:
            self.lga_net_params = load_pkl_object(net_ckpt_path)
            print(f"Loaded LGA model from ckpt: {net_ckpt_path}")

        if net_params is None and net_ckpt_path is None:
            if self.num_dims > 50:
                ckpt_fname = "2023_04_lga_v7.pkl"
            else:
                ckpt_fname = "2023_04_lga_v4.pkl"
            data = pkgutil.get_data(__name__, f"ckpt/lga/{ckpt_fname}")
            self.lga_net_params = load_pkl_object(data, pkg_load=True)
            print(f"Loaded pretrained LGA model from ckpt: {ckpt_fname}")

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(net_params=self.lga_net_params, sigma_init=self.sigma_init)

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """`initialize` the evolution strategy."""
        init_x = jax.random.uniform(
            rng,
            (self.elite_popsize, self.num_dims),
            minval=params.init_min,
            maxval=params.init_max,
        )
        init_sigma = jnp.ones((self.elite_popsize, 1)) * params.sigma_init

        return EvoState(
            rng=rng,
            mean=init_x[0],
            archive_x=init_x,
            archive_f=jnp.zeros(self.elite_popsize) + 5000000.0,
            archive_sigma=init_sigma,
            archive_age=jnp.zeros(self.elite_popsize),
            sigma_C=jnp.zeros((self.popsize, 1)),
            best_member=init_x[0],
        )

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        rng, rng_idx = jax.random.split(rng)
        elite_ids = jnp.arange(self.elite_popsize)

        # Sample candidates with replacement given distribution
        # Get probabilities to sample children from parent archive
        age_features = tanh_age(state.archive_age, state.gen_counter + 1e-10)
        F_E = self.fitness_features.apply(
            state.archive_x, state.archive_f, state.best_fitness
        )
        F_E = jnp.concatenate([F_E, age_features.reshape(-1, 1)], axis=1)
        p = self.sampling_layer.apply(params.net_params["sampling"], F_E)
        idx = jax.random.choice(rng_idx, elite_ids, (self.popsize,), p=p)

        # Select children with sampled indices
        X_C = state.archive_x[idx]
        f_C = state.archive_f[idx]
        sigma_C = state.archive_sigma[idx]

        # Perform mutation rate adaptation of solutions
        F_C_tilde = self.fitness_features.apply(X_C, f_C, state.best_fitness)
        sigma_C = self.mutation_layer.apply(
            params.net_params["mutation"], sigma_C, F_C_tilde
        )

        # Perform Gaussian scaled mutation
        epsilon = jax.random.normal(rng, (self.popsize, self.num_dims))
        x = X_C + sigma_C * epsilon
        return x, state.replace(sigma_C=sigma_C)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        fit_all = jnp.concatenate([fitness, state.archive_f])
        x_all = jnp.concatenate([x, state.archive_x])
        fit_re = self.fitness_features.apply(x_all, fit_all, state.best_fitness)
        idx = jnp.argsort(fit_all)[: self.elite_popsize]

        rng, rng_next = jax.random.split(state.rng)
        # Perform selection - either learned or hard truncation based
        F_X, F_E = fit_re[: self.popsize], fit_re[self.popsize :]
        select_bool = self.selection_layer.apply(
            params.net_params["selection"], rng, F_X, F_E
        )
        keep_parent = (select_bool.sum(axis=1) == 0)[:, None]
        next_x = select_bool @ x + keep_parent * state.archive_x
        next_f = select_bool @ fitness + keep_parent.squeeze() * state.archive_f
        next_sigma = select_bool @ state.sigma_C + keep_parent * state.archive_sigma

        # Update the age counter - reset if copy over otherwise increase
        next_age = state.archive_age * keep_parent.squeeze() + keep_parent.squeeze()

        # Argsort by performance and set mean
        improved = fit_all[idx][0] < state.best_fitness
        best_mean = jax.lax.select(improved, x_all[idx][0], state.best_member)
        return state.replace(
            rng=rng_next,
            mean=best_mean,
            archive_x=next_x,
            archive_f=next_f,
            archive_sigma=next_sigma,
            archive_age=next_age,
        )



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    archive: chex.Array
    fitness: chex.Array
    velocity: chex.Array
    best_archive: chex.Array
    best_archive_fitness: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    inertia_coeff: float = 0.75  # w momentum of velocity
    cognitive_coeff: float = 1.5  # c_1 cognitive "force" multiplier
    social_coeff: float = 2.0  # c_2 social "force" multiplier
    init_min: float = -0.1
    init_max: float = 0.1
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class PSO(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Particle Swarm Optimization (Kennedy & Eberhart, 1995)
        Reference: https://ieeexplore.ieee.org/document/488968"""
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            n_devices=n_devices,
            **fitness_kwargs
        )
        self.strategy_name = "PSO"

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams()

    def initialize_strategy(
        self, rng: chex.PRNGKey, params: EvoParams
    ) -> EvoState:
        """`initialize` the evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.popsize, self.num_dims),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization.mean(axis=0),
            archive=initialization,
            fitness=jnp.zeros(self.popsize) + jnp.finfo(jnp.float32).max,
            velocity=jnp.zeros((self.popsize, self.num_dims)),
            best_archive=initialization,
            best_archive_fitness=jnp.zeros(self.popsize)
            + jnp.finfo(jnp.float32).max,
            best_member=initialization.mean(axis=0),
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """
        `ask` for new proposed candidates to evaluate next.
        1. Update v_i(t+1) velocities base on:
          - Inertia: w * v_i(t)
          - Cognitive: c_1 * r_1 * (p_(i, lb)(t) - x_i(t))
          - Social: c_2 * r_2 * (p_(gb)(t) - x_i(t))
        2. Update "particle" positions: x_i(t+1) = x_i(t) + v_i(t+1)
        """
        rng_members = jax.random.split(rng, self.popsize)
        member_ids = jnp.arange(self.popsize)
        vel = jax.vmap(
            single_member_velocity,
            in_axes=(0, 0, None, None, None, None, None, None, None),
        )(
            rng_members,
            member_ids,
            state.archive,
            state.velocity,
            state.best_archive,
            state.best_archive_fitness,
            params.inertia_coeff,
            params.cognitive_coeff,
            params.social_coeff,
        )
        # Update particle positions with velocity
        x = state.archive + vel
        return x, state.replace(velocity=vel)

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """
        `tell` update to ES state.
        If fitness of y <= fitness of x -> replace in population.
        """
        replace = fitness <= state.best_archive_fitness
        best_archive = (
            jnp.expand_dims(replace, 1) * x
            + (1 - jnp.expand_dims(replace, 1)) * state.best_archive
        )
        best_archive_fitness = (
            replace * fitness + (1 - replace) * state.best_archive_fitness
        )
        return state.replace(
            mean=x.mean(axis=0),
            fitness=fitness,
            archive=x,
            best_archive=best_archive,
            best_archive_fitness=best_archive_fitness,
        )


def single_member_velocity(
    rng: chex.PRNGKey,
    member_id: int,
    archive: chex.Array,
    velocity: chex.Array,
    best_archive: chex.Array,
    best_fitness: chex.Array,
    inertia_coeff: float,
    cognitive_coeff: float,
    social_coeff: float,
):
    """Update v_i(t+1) velocities based on: Inertia, Cognitive, Social."""
    # Sampling one shared r1, r2 across dims of one member seems more robust!
    # r1, r2 = jax.random.uniform(rng, (2, archive.shape[1]))
    r1, r2 = jax.random.uniform(rng, (2,))
    global_best_id = jnp.argmin(best_fitness)
    global_best = best_archive[global_best_id]
    vel_new = (
        inertia_coeff * velocity[member_id]
        + cognitive_coeff * r1 * (best_archive[member_id] - archive[member_id])
        + social_coeff * r2 * (global_best - archive[member_id])
    )
    return vel_new



import jax
import jax.numpy as jnp
from evosax.networks import LSTM, MLP
from evosax.experimental.decodings import RandomDecoder, HyperDecoder


# def test_random_decoding():
#     rng = jax.random.PRNGKey(0)
#     num_encoding_dims = 10
#     popsize = 100
#     # Define some network architecture to optimize params for
#     network = LSTM(
#         num_hidden_units=10,
#         num_output_units=1,
#         output_activation="identity",
#     )
#     net_params = network.init(
#         rng,
#         x=jnp.zeros((10,)),
#         carry=network.initialize_carry(),
#         rng=rng,
#     )

#     # Define candidates coming from strategy.ask - batch reshape
#     encodings = jnp.zeros((popsize, num_encoding_dims))
#     decoder = RandomDecoder(num_encoding_dims, net_params)
#     x = decoder.reshape(encodings)
#     assert x["params"]["LSTMCell_0"]["hf"]["kernel"].shape == (popsize, 10, 10)

#     # Define candidates coming from strategy.ask - single reshape
#     encodings = jnp.zeros(num_encoding_dims)
#     decoder = RandomDecoder(num_encoding_dims, net_params)
#     x = decoder.reshape_single(encodings)
#     assert x["params"]["LSTMCell_0"]["hf"]["kernel"].shape == (10, 10)


def test_hypermlp_decoding():
    rng = jax.random.PRNGKey(0)
    popsize = 100
    # Define some network architecture to optimize params for
    network = MLP(
        num_hidden_units=64,
        num_hidden_layers=2,
        num_output_units=1,
        output_activation="identity",
    )
    net_params = network.init(
        rng,
        x=jnp.zeros((10,)),
        rng=rng,
    )

    # Define candidates coming from strategy.ask - batch reshape
    decoder = HyperDecoder(
        net_params,
        hypernet_config={
            "num_latent_units": 3,  # Latent units per module kernel/bias
            "num_hidden_units": 2,  # Hidden dimensionality of a_i^j embedding
        },
    )
    encodings = jnp.zeros((popsize, decoder.num_encoding_dims))
    x = decoder.reshape(encodings)
    assert x["params"]["Dense_0"]["kernel"].shape == (popsize, 10, 64)

    # # Define candidates coming from strategy.ask - single reshape
    # encodings = jnp.zeros(decoder.num_encoding_dims)
    # x = decoder.reshape_single(encodings)
    # assert x["params"]["Dense_0"]["kernel"].shape == (10, 64)



import jax
import jax.numpy as jnp
import chex
from typing import Optional, Dict, Tuple
from flax import struct

# TODO: Add gradient clipping - select leads to more compute
# "use_clip_by_global_norm": False,
# "clip_global_norm": 5,
# "use_clip_by_value": False,
# "clip_value": 5,


def exp_decay(
    param: chex.Array, param_decay: chex.Array, param_limit: chex.Array
) -> chex.Array:
    """Exponentially decay parameter & clip by minimal value."""
    param = param * param_decay
    param = jnp.maximum(param, param_limit)
    return param


@struct.dataclass
class OptState:
    lrate: float
    m: chex.Array
    v: Optional[chex.Array] = None
    n: Optional[chex.Array] = None
    last_grads: Optional[chex.Array] = None
    gen_counter: int = 0


@struct.dataclass
class OptParams:
    lrate_init: float = 0.01
    lrate_decay: float = 0.999
    lrate_limit: float = 0.001
    momentum: Optional[float] = None
    beta_1: Optional[float] = None
    beta_2: Optional[float] = None
    beta_3: Optional[float] = None
    eps: Optional[float] = None
    max_speed: Optional[float] = None


class Optimizer(object):
    def __init__(self, num_dims: int):
        """Simple JAX-Compatible Optimizer Class."""
        self.num_dims = num_dims

    @property
    def default_params(self) -> OptParams:
        """Return shared and optimizer-specific default parameters."""
        return OptParams(**self.params_opt)

    def initialize(self, params: OptParams) -> OptState:
        """Initialize the optimizer state."""
        return self.initialize_opt(params)

    def step(
        self,
        mean: chex.Array,
        grads: chex.Array,
        state: OptState,
        params: OptParams,
    ) -> [chex.Array, OptState]:
        """Perform a gradient-based update step."""
        return self.step_opt(mean, grads, state, params)

    def update(self, state: OptState, params: OptParams) -> OptState:
        """Exponentially decay the learning rate if desired."""
        lrate = exp_decay(state.lrate, params.lrate_decay, params.lrate_limit)
        return state.replace(lrate=lrate)

    @property
    def params_opt(self) -> OptParams:
        """Optimizer-specific hyperparameters."""
        raise NotImplementedError

    def initialize_opt(self, params: OptParams) -> OptState:
        """Optimizer-specific initialization of optimizer state."""
        raise NotImplementedError

    def step_opt(
        self,
        mean: chex.Array,
        grads: chex.Array,
        state: OptState,
        params: OptParams,
    ) -> Tuple[chex.Array, OptState]:
        """Optimizer-specific step to update parameter estimates."""
        raise NotImplementedError


class SGD(Optimizer):
    def __init__(self, num_dims: int):
        """Simple JAX-Compatible SGD + Momentum optimizer."""
        super().__init__(num_dims)
        self.opt_name = "sgd"

    @property
    def params_opt(self) -> Dict[str, float]:
        """Return default SGD+Momentum parameters."""
        return {
            "momentum": 0.0,
        }

    def initialize_opt(self, params: OptParams) -> OptState:
        """Initialize the momentum trace of the optimizer."""
        return OptState(m=jnp.zeros(self.num_dims), lrate=params.lrate_init)

    def step_opt(
        self,
        mean: chex.Array,
        grads: chex.Array,
        state: chex.ArrayTree,
        params: chex.ArrayTree,
    ) -> Tuple[chex.Array, OptState]:
        """Perform a simple SGD + Momentum step."""
        m = grads + params.momentum * state.m
        mean_new = mean - state.lrate * state.m
        return mean_new, state.replace(m=m, gen_counter=state.gen_counter + 1)


class Adam(Optimizer):
    def __init__(self, num_dims: int):
        """JAX-Compatible Adam Optimizer (Kingma & Ba, 2015)
        Reference: https://arxiv.org/abs/1412.6980"""
        super().__init__(num_dims)
        self.opt_name = "adam"

    @property
    def params_opt(self) -> Dict[str, float]:
        """Return default Adam parameters."""
        return {
            "beta_1": 0.99,
            "beta_2": 0.999,
            "eps": 1e-8,
        }

    def initialize_opt(self, params: OptParams) -> OptState:
        """Initialize the m, v trace of the optimizer."""
        return OptState(
            m=jnp.zeros(self.num_dims),
            v=jnp.zeros(self.num_dims),
            lrate=params.lrate_init,
        )

    def step_opt(
        self,
        mean: chex.Array,
        grads: chex.Array,
        state: OptState,
        params: OptParams,
    ) -> Tuple[chex.Array, OptState]:
        """Perform a simple Adam GD step."""
        m = (1 - params.beta_1) * grads + params.beta_1 * state.m
        v = (1 - params.beta_2) * (grads ** 2) + params.beta_2 * state.v
        mhat = m / (1 - params.beta_1 ** (state.gen_counter + 1))
        vhat = v / (1 - params.beta_2 ** (state.gen_counter + 1))
        mean_new = mean - state.lrate * mhat / (jnp.sqrt(vhat) + params.eps)
        return mean_new, state.replace(
            m=m, v=v, gen_counter=state.gen_counter + 1
        )


class RMSProp(Optimizer):
    def __init__(self, num_dims: int):
        """JAX-Compatible RMSProp Optimizer (Hinton et al., 2012)
        Reference: https://tinyurl.com/2sbbcnrv"""
        super().__init__(num_dims)
        self.opt_name = "rmsprop"

    @property
    def params_opt(self) -> Dict[str, float]:
        """Return default RMSProp parameters."""
        return {
            "momentum": 0.9,
            "beta_1": 0.99,
            "eps": 1e-8,
        }

    def initialize_opt(self, params: OptParams) -> OptState:
        """Initialize the m, v trace of the optimizer."""
        return OptState(
            m=jnp.zeros(self.num_dims),
            v=jnp.zeros(self.num_dims),
            lrate=params.lrate_init,
        )

    def step_opt(
        self,
        mean: chex.Array,
        grads: chex.Array,
        state: OptState,
        params: OptParams,
    ) -> Tuple[chex.Array, OptState]:
        """Perform a simple RMSprop GD step."""
        v = (1 - params.beta_1) * (grads ** 2) + params.beta_1 * state.v
        m = params.momentum * state.m + grads / (jnp.sqrt(v) + params.eps)
        mean_new = mean - state.lrate * m
        return mean_new, state.replace(
            m=m, v=v, gen_counter=state.gen_counter + 1
        )


class ClipUp(Optimizer):
    def __init__(self, num_dims: int):
        """JAX-Compatible ClipUp Optimizer (Toklu et al., 2020)
        Reference: https://arxiv.org/abs/2008.02387"""
        super().__init__(num_dims)
        self.opt_name = "clipup"

    @property
    def params_opt(self) -> Dict[str, float]:
        """Return default ClipUp parameters."""
        return {
            "lrate_init": 0.15,
            "lrate_decay": 0.999,
            "lrate_limit": 0.05,
            "max_speed": 0.3,
            "momentum": 0.9,
        }

    def initialize_opt(self, params: OptParams) -> OptState:
        """Initialize the momentum trace of the optimizer."""
        return OptState(m=jnp.zeros(self.num_dims), lrate=params.lrate_init)

    def step_opt(
        self,
        mean: chex.Array,
        grads: chex.Array,
        state: OptState,
        params: OptParams,
    ) -> Tuple[chex.Array, OptState]:
        """Perform a ClipUp step. mom = 0.9, lrate = vmax/2, vmax = small"""
        # Normalize length of gradients - vmax & alpha control max step magnitude
        grad_magnitude = jnp.sqrt(jnp.sum(grads * grads))
        gradient = grads / (grad_magnitude + 1e-08)
        step = gradient * state.lrate
        velocity = params.momentum * state.m + step

        def clip(velocity: chex.Array, max_speed: float):
            """Rescale clipped velocities."""
            vel_magnitude = jnp.sqrt(jnp.sum(velocity * velocity))
            ratio_scale = vel_magnitude > max_speed
            scaled_vel = velocity * (max_speed / (vel_magnitude + 1e-08))
            x_out = jax.lax.select(ratio_scale, scaled_vel, velocity)
            return x_out

        # Clip the update velocity and apply the update
        m = clip(velocity, params.max_speed)
        mean_new = mean - state.lrate * m
        return mean_new, state.replace(m=m)


class Adan(Optimizer):
    def __init__(self, num_dims: int):
        """JAX-Compatible Adan Optimizer (Xi et al., 2022)
        Reference: https://arxiv.org/pdf/2208.06677.pdf"""
        super().__init__(num_dims)
        self.opt_name = "adan"

    @property
    def params_opt(self) -> Dict[str, float]:
        """Return default Adam parameters."""
        return {
            "beta_1": 0.98,
            "beta_2": 0.92,
            "beta_3": 0.99,
            "eps": 1e-8,
        }

    def initialize_opt(self, params: OptParams) -> OptState:
        """Initialize the m, v, n trace of the optimizer."""
        return OptState(
            m=jnp.zeros(self.num_dims),
            v=jnp.zeros(self.num_dims),
            n=jnp.zeros(self.num_dims),
            last_grads=jnp.zeros(self.num_dims),
            lrate=params.lrate_init,
        )

    def step_opt(
        self,
        mean: chex.Array,
        grads: chex.Array,
        state: OptState,
        params: OptParams,
    ) -> Tuple[chex.Array, OptState]:
        """Perform a simple Adan GD step."""
        m = (1 - params.beta_1) * grads + params.beta_1 * state.m
        grad_diff = grads - state.last_grads
        v = (1 - params.beta_2) * grad_diff + params.beta_2 * state.v
        n = (1 - params.beta_3) * (
            grads + params.beta_2 * grad_diff
        ) ** 2 + params.beta_3 * state.n

        mhat = m / (1 - params.beta_1 ** (state.gen_counter + 1))
        vhat = v / (1 - params.beta_2 ** (state.gen_counter + 1))
        nhat = n / (1 - params.beta_3 ** (state.gen_counter + 1))
        mean_new = mean - state.lrate * (mhat + params.beta_2 * vhat) / (
            jnp.sqrt(nhat) + params.eps
        )
        return mean_new, state.replace(
            m=m, v=v, n=n, last_grads=grads, gen_counter=state.gen_counter + 1
        )



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from flax import linen as nn
from ..strategy import Strategy


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: chex.Array
    weights: chex.Array  # Weights for population members
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    temperature: float = 12.5  # Temperature for softmax weights
    lrate_sigma: float = 0.1  # Learning rate for population std
    lrate_mean: float = 1.0  # Learning rate for population mean
    sigma_init: float = 0.1  # Standard deviation
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


def get_des_weights(popsize: int, temperature: float = 12.5):
    """Compute discovered recombination weights."""
    ranks = jnp.arange(popsize)
    ranks /= ranks.size - 1
    ranks = ranks - 0.5
    sigout = nn.sigmoid(temperature * ranks)
    weights = nn.softmax(-20 * sigout)
    return weights[:, None]


class DES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        temperature: float = 12.5,
        sigma_init: float = 0.1,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Discovered Evolution Strategy (Lange et al., 2023)"""
        super().__init__(
            popsize, num_dims, pholder_params, mean_decay, n_devices, **fitness_kwargs
        )
        self.strategy_name = "DES"
        self.temperature = temperature
        self.sigma_init = sigma_init

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams(temperature=self.temperature, sigma_init=self.sigma_init)

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """`initialize` the evolution strategy."""
        # Get DES discovered recombination weights.
        weights = get_des_weights(self.popsize, params.temperature)
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init * jnp.ones(self.num_dims),
            weights=weights,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        z = jax.random.normal(rng, (self.popsize, self.num_dims))  # ~ N(0, I)
        x = state.mean + z * state.sigma.reshape(1, self.num_dims)  # ~ N(m, σ^2 I)
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        weights = state.weights
        x = x[fitness.argsort()]
        # Weighted updates
        weighted_mean = (weights * x).sum(axis=0)
        weighted_sigma = jnp.sqrt((weights * (x - state.mean) ** 2).sum(axis=0) + 1e-06)
        mean = state.mean + params.lrate_mean * (weighted_mean - state.mean)
        sigma = state.sigma + params.lrate_sigma * (weighted_sigma - state.sigma)
        return state.replace(mean=mean, sigma=sigma)



"""Implementation of BBOB functions provided in Hansen et al. (2009)."""
from typing import Optional
import chex
import jax
import jax.numpy as jnp
from functools import partial
from evosax.utils.visualizer_2d import BBOBVisualizer
from .bbob_helpers import (
    get_rotation,
    lambda_alpha_trafo,
    oscillation_trafo,
    asymmetry_trafo,
    boundary_penalty,
)


class BBOBFitness(object):
    """BBOB Functions Benchmark Task.
    Functions from Hansen et al. (2009)
    'Real-parameter black-box optimization benchmarking 2009: Noiseless functions definitions'
    Link: https://numbbo.github.io/gforge/downloads/download16.00/bbobdocfunctions.pdf
    """

    def __init__(
        self,
        fn_name: str = "Sphere",
        num_dims: int = 2,
        seed_id: int = 1,
    ):
        self.num_dims = num_dims
        # Create rotation matrices for non-separability
        rng = jax.random.PRNGKey(seed_id)
        self.seed_id = seed_id
        self.R, self.Q = self.get_rotation_matrices(rng)
        self.fn_name = fn_name
        self.fn = BBOB_fns[self.fn_name]

    @partial(jax.jit, static_argnums=(0,))
    def rollout(
        self,
        rng: chex.PRNGKey,
        eval_params: chex.Array,
        R: Optional[chex.Array] = None,
        Q: Optional[chex.Array] = None,
        noise_std: float = 0.0,
    ) -> chex.Array:
        """Batch evaluate the proposal points."""
        if R is None:
            R, Q = self.R, self.Q
        val = jax.vmap(self.fn, in_axes=(0, None, None))(eval_params, R, Q)
        eval_noise = jax.random.normal(rng, shape=val.shape) * noise_std
        return val + eval_noise

    def get_rotation_matrices(self, rng: chex.PRNGKey):
        """Sample two rotation matrices."""
        rng_q, rng_r = jax.random.split(rng)
        R = get_rotation(rng_r, self.num_dims)
        Q = get_rotation(rng_q, self.num_dims)
        return R, Q

    def visualize(self, plot_log_fn: bool = False):
        """Visualize fitness landscape in 2D case."""
        assert self.num_dims == 2
        visualizer = BBOBVisualizer(
            None, None, self.fn_name, self.fn_name, True, plot_log_fn
        )
        visualizer.plot_contour_3d(save=False)


def Sphere(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Sphere fn - separable (p.5; Hansen et al., 2009)."""
    out = arr * arr
    return jnp.sum(out)


def EllipsoidalOriginal(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Ellipsoidal Original fn - separable (p.10; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z_vec = jax.vmap(oscillation_trafo)(arr)

    def get_val(i):
        exp = jax.lax.select(dim > 1, 6.0 * i / (dim - 1), 6.0)
        s = 10 ** exp * z_vec[i] * z_vec[i]
        return s

    out = jax.vmap(get_val)(jnp.arange(dim))
    return jnp.sum(out)


def RastriginOriginal(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Rastrigin fn - separable (p.15; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z = asymmetry_trafo(jax.vmap(oscillation_trafo)(arr), 0.2, dim)
    z = jnp.matmul(lambda_alpha_trafo(10.0, dim), z)
    out_1 = jnp.cos(2 * jnp.pi * z)
    out_2 = z * z
    return 10 * (dim - jnp.sum(out_1)) + jnp.sum(out_2)


def BuecheRastrigin(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Bueche-Rastrigin fn - separable (p.20; Hansen et al., 2009)."""
    dim = arr.shape[0]
    t = jax.vmap(oscillation_trafo)(arr)

    def calculate_si(
        max_dims: int, vector: chex.Array, num_dims: int
    ) -> chex.Array:
        def get_val(i):
            s = jax.lax.select(
                num_dims > 1, 10 ** (0.5 * (i / (num_dims - 1.0))), 10 ** 0.5
            )
            extra_cond = jnp.logical_and(i % 2 == 0, vector[i] > 0)
            s = jax.lax.select(extra_cond, s * 10, s)
            return s

        return jax.vmap(get_val, in_axes=0)(jnp.arange(max_dims))

    L = calculate_si(dim, arr, dim) * t
    out1 = jnp.cos(2 * jnp.pi * L)
    out2 = L * L
    t1 = 10 * (dim - jnp.sum(out1 * dim))
    t2 = jnp.sum(out2 * dim)
    t3 = 100 * boundary_penalty(arr, dim)
    return t1 + t2 + t3


def LinearSlope(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Linear Slope fn - separable (p.25; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z = jnp.matmul(R, arr)

    def get_val(i):
        s = 10 ** jax.lax.select(dim > 1, i / (dim - 1.0), 1.0)
        z_opt = 5 * jnp.sum(jnp.abs(R[i, :]))
        return s * (z_opt - z[i])

    out = jax.vmap(get_val)(jnp.arange(dim))
    return jnp.sum(out)


def AttractiveSector(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Attractive Sector fn - low/mod condition (p.30; Hansen et al., 2009)."""
    dim = arr.shape[0]
    x_opt = jnp.array([1 * (i % 2 == 0) - 1 * (i % 2 != 0) for i in range(dim)])
    z_vec = jnp.matmul(R, arr - x_opt)
    z_vec = jnp.matmul(lambda_alpha_trafo(10.0, dim), z_vec)
    z_vec = jnp.matmul(Q, z_vec)

    def get_val(i):
        z = z_vec[i]
        s = jax.lax.select(z * x_opt[i] > 0, 100, 1)
        return (s * z) ** 2

    out = jax.vmap(get_val)(jnp.arange(dim))
    res = jnp.sum(out)
    return jnp.power(oscillation_trafo(res), 0.9)


def StepEllipsoidal(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Step Ellipsoidal fn - low/mod condition (p.35; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z_hat = jnp.matmul(R, arr)
    z_hat = jnp.matmul(lambda_alpha_trafo(10.0, dim), z_hat)
    z_tilde = jnp.array(
        [
            jnp.floor(0.5 + z) * (z > 0.5)
            + (jnp.floor(0.5 + 10 * z)) * (z <= 0.5)
            for z in z_hat
        ]
    )
    z_tilde = jnp.matmul(Q, z_tilde)

    def get_val(i):
        exponent = jax.lax.select(dim > 1.0, 2.0 * i / (dim - 1.0), 2.0)
        return 10.0 ** exponent * z_tilde[i] ** 2

    out = jax.vmap(get_val)(jnp.arange(dim))
    res = jnp.sum(out)
    value = jnp.maximum(jnp.abs(z_hat[0]) / 1000, res)
    return 0.1 * value + boundary_penalty(arr, dim)


def RosenbrockOriginal(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """RosenbrockOriginal fn - low/mod condition (p.40; Hansen et al., 2009)."""
    dim = arr.shape[0]
    # Optimum is at origin!
    z = jnp.maximum(1.0, (dim ** 0.5) / 8.0) * arr + 1
    out = jnp.array(
        [
            100.0 * (z[i] ** 2 - z[i + 1]) ** 2 + (z[i] - 1) ** 2
            for i in range(dim - 1)
        ]
    )
    return jnp.sum(out)


def RosenbrockRotated(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """RosenbrockRotated fn - low/mod condition (p.45; Hansen et al., 2009)."""
    dim = arr.shape[0]
    r_x = jnp.matmul(R, arr)
    z = jnp.maximum(1.0, (dim ** 0.5) / 8.0) * r_x + 0.5 * jnp.ones((dim,))
    out = jnp.array(
        [
            100.0 * (z[i] ** 2 - z[i + 1]) ** 2 + (z[i] - 1) ** 2
            for i in range(dim - 1)
        ]
    )
    return jnp.sum(out)


def EllipsoidalRotated(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Ellipsoidal Rotated fn - high condition (p.50; Hansen et al., 2009)."""
    dim = arr.shape[0]
    r_x = jnp.matmul(R, arr)
    z_vec = jax.vmap(oscillation_trafo)(r_x)

    def get_val(i):
        exp = jax.lax.select(dim > 1, 6.0 * i / (dim - 1), 6.0)
        s = 10 ** exp * z_vec[i] * z_vec[i]
        return s

    out = jax.vmap(get_val)(jnp.arange(dim))
    return jnp.sum(out)


def Discus(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Discus fn - high condition (p.55; Hansen et al., 2009)."""
    r_x = jnp.matmul(R, arr)
    z_vec = jax.vmap(oscillation_trafo)(r_x)
    out = jnp.array([z * z for z in z_vec[1:]])
    return 10 ** 6 * z_vec[0] * z_vec[0] + jnp.sum(out)


def BentCigar(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Bent Cigar fn - high condition (p.60; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z_vec = jnp.matmul(R, arr)
    z_vec = asymmetry_trafo(z_vec, 0.5, dim)
    z_vec = jnp.matmul(R, z_vec)
    out = z_vec[1:] ** 2
    return z_vec[0] ** 2 + 10 ** 6 * jnp.sum(out)


def SharpRidge(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Sharp Ridge fn - high condition (p.65; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z_vec = jnp.matmul(R, arr)
    z_vec = jnp.matmul(lambda_alpha_trafo(10, dim), z_vec)
    z_vec = jnp.matmul(Q, z_vec)
    out = z_vec[1:] ** 2
    return z_vec[0] ** 2 + 100 * jnp.sum(out) ** 0.5


def DifferentPowers(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Different Powers fn - high condition (p.70; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z = jnp.matmul(R, arr)

    def get_val(i):
        exp = (2 + 4 * i / (dim - 1)) * (dim > 1) + (dim <= 1) * 6
        s = jnp.abs(z[i]) ** exp
        return s

    out = jax.vmap(get_val)(jnp.arange(dim))
    return jnp.sum(out) ** 0.5


def RastriginRotated(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Rastrigin fn - multi-modal (p.75; Hansen et al., 2009)."""
    dim = arr.shape[0]
    z = jnp.matmul(R, arr)
    z = asymmetry_trafo(jax.vmap(oscillation_trafo)(z), 0.2, dim)
    z = jnp.matmul(Q, z)
    z = jnp.matmul(lambda_alpha_trafo(10.0, dim), z)
    z = jnp.matmul(R, z)
    out_1 = jnp.cos(2 * jnp.pi * z)
    out_2 = z * z
    return 10 * (dim - jnp.sum(out_1)) + jnp.sum(out_2)


def Weierstrass(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Weierstrass fn - multi-modal (p.80; Hansen et al., 2009)."""
    k_order = 12
    dim = arr.shape[0]
    z = jnp.matmul(R, arr)
    z = jax.vmap(oscillation_trafo)(z)
    z = jnp.matmul(Q, z)
    z = jnp.matmul(lambda_alpha_trafo(1.0 / 100.0, dim), z)
    f0 = jnp.array(
        [0.5 ** k * jnp.cos(jnp.pi * 3 ** k) for k in range(k_order)]
    ).sum()

    def get_val(i):
        s = 0
        for k in range(k_order):
            s += 0.5 ** k * jnp.cos(2 * jnp.pi * (3 ** k) * (z[i] + 0.5))
        return s

    out = jax.vmap(get_val)(jnp.arange(dim))
    s = jnp.sum(out)
    return 10 * (s / dim - f0) ** 3 + 10 * boundary_penalty(arr, dim) / dim


def SchaffersF7(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """SchaffersF7 fn - multi-modal (p.85; Hansen et al., 2009)."""
    dim = arr.shape[0]
    if dim == 1:
        return 0.0
    z = jnp.matmul(R, arr)
    z = asymmetry_trafo(z, 0.5, dim)
    z = jnp.matmul(Q, z)
    z = jnp.matmul(lambda_alpha_trafo(10.0, dim), z)

    def get_val_arr(i):
        s_arr = (z[i] ** 2 + z[i + 1] ** 2) ** 0.5
        return s_arr

    s_arr = jax.vmap(get_val_arr, in_axes=0)(jnp.arange(dim - 1))

    def get_val(i):
        return (
            s_arr[i] ** 0.5
            + (s_arr[i] ** 0.5) * jnp.sin(50 * s_arr[i] ** 0.2) ** 2
        )

    out = jax.vmap(get_val)(jnp.arange(dim - 1))
    s = jnp.sum(out)
    return (s / (dim - 1.0)) ** 2 + 10 * boundary_penalty(arr, dim)


def SchaffersF7IllConditioned(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """SchaffersF7 ill condition - multi-modal (p.90; Hansen et al., 2009)."""
    dim = arr.shape[0]
    if dim == 1:
        return 0.0
    z = jnp.matmul(R, arr)
    z = asymmetry_trafo(z, 0.5, dim)
    z = jnp.matmul(Q, z)
    z = jnp.matmul(lambda_alpha_trafo(1000.0, dim), z)

    def get_val_arr(i):
        s_arr = (z[i] ** 2 + z[i + 1] ** 2) ** 0.5
        return s_arr

    s_arr = jax.vmap(get_val_arr, in_axes=0)(jnp.arange(dim - 1))

    def get_val(i):
        return (
            s_arr[i] ** 0.5
            + (s_arr[i] ** 0.5) * jnp.sin(50 * s_arr[i] ** 0.2) ** 2
        )

    out = jax.vmap(get_val)(jnp.arange(dim - 1))
    s = jnp.sum(out)
    return (s / (dim - 1.0)) ** 2 + 10 * boundary_penalty(arr, dim)


def GriewankRosenbrock(
    arr: chex.Array, R: chex.Array, Q: chex.Array
) -> chex.Array:
    """Griewank-Rosenbrock fn - multi-modal weak (p.95; Hansen et al., 2009)."""
    dim = arr.shape[0]
    r_x = jnp.matmul(R, arr)
    # Original BBOB: max(1.0, (dim ** 0.5)/8) * r_x + 0.5 * jnp.ones((dim,))
    z_arr = jnp.maximum(1.0, (dim ** 0.5) / 8.0) * r_x + jnp.ones((dim,))
    s_arr = jnp.zeros(dim)

    def get_val_arr(i):
        s_arr = (
            100.0 * (z_arr[i] ** 2 - z_arr[i + 1]) ** 2 + (z_arr[i] - 1) ** 2
        )
        return s_arr

    s_arr = jax.vmap(get_val_arr)(jnp.arange(dim - 1))

    def get_val(i):
        val = s_arr[i] / 4000.0 - jnp.cos(s_arr[i])
        return val

    out = jax.vmap(get_val)(jnp.arange(dim - 1))
    total = jnp.sum(out)
    return (10.0 * total) / (dim - 1) + 10


def Schwefel(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Schwefel fn - multi-modal weak (p.100; Hansen et al., 2009)."""
    dim = arr.shape[0]
    bernoulli_arr = jnp.array([jnp.power(-1, i + 1) for i in range(dim)])
    x_opt = 4.2096874633 / 2.0 * bernoulli_arr
    x_hat = 2.0 * (bernoulli_arr * arr)  # element-wise multiply
    z_hat = jnp.zeros(dim)
    z_hat = z_hat.at[0].set(x_hat[0])

    def get_val(i):
        val = x_hat[i] + 0.25 * (x_hat[i - 1] - 2 * jnp.abs(x_opt[i - 1]))
        return val

    vals = jax.vmap(get_val)(jnp.arange(1, dim))
    z_hat = z_hat.at[1:].set(vals)
    z_vec = 100 * (
        jnp.matmul(lambda_alpha_trafo(10, dim), z_hat - 2 * jnp.abs(x_opt))
        + 2 * jnp.abs(x_opt)
    )
    out = z_vec * jnp.sin(jnp.abs(z_vec) ** 0.5)
    total = jnp.sum(out)
    return (
        -(total / (100.0 * dim))
        + 4.189828872724339
        + 100 * boundary_penalty(z_vec / 100, dim)
    )


def Gallagher101Me(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Gallagher 101 peaks - multi-modal weak (p.105; Hansen et al., 2009)."""
    dim = arr.shape[0]
    num_optima = 101

    def get_optima(i):
        vec = jnp.zeros(dim)
        for j in range(dim):
            alpha = ((i - 1) * dim + j + 1.0) / (dim * num_optima + 2.0)
            vec = vec.at[j].set(-5 + 10 * alpha)
        return vec * (i != 0) + jnp.zeros([dim]) * (i == 0)

    optima_list = jax.vmap(get_optima)(jnp.arange(num_optima))

    def get_c_val(i):
        alpha = 1000.0 ** (2.0 * (i - 1) / (num_optima - 2))
        c_mat = lambda_alpha_trafo(alpha, dim) / (alpha ** 0.25)
        return c_mat * (i != 0) + lambda_alpha_trafo(1000, dim) * (i == 0)

    c_list = jax.vmap(get_c_val)(jnp.arange(num_optima))
    max_value = jnp.array([-1.0])

    for i in range(num_optima):
        w = 10 * (i == 0) + (1.1 + 8.0 * (i - 1.0) / (num_optima - 2.0)) * (
            i != 0
        )
        diff = jnp.matmul(R, (arr - optima_list[i]).reshape(-1, 1))
        e = jnp.matmul(diff.transpose(), jnp.matmul(c_list[i], diff))
        max_value = jnp.maximum(max_value, w * jnp.exp(-e / (2.0 * dim)))
    return oscillation_trafo(
        10.0 - max_value.squeeze()
    ) ** 2 + boundary_penalty(arr, dim)


def Gallagher21Hi(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Gallagher 21 peaks fn - multi-modal weak (p.110; Hansen et al., 2009)."""
    dim = arr.shape[0]
    num_optima = 21

    def get_optima(i):
        vec = jnp.zeros(dim)
        for j in range(dim):
            alpha = ((i - 1) * dim + j + 1.0) / (dim * num_optima + 2.0)
            vec = vec.at[j].set(-5 + 10 * alpha)
        return vec * (i != 0) + jnp.zeros([dim]) * (i == 0)

    optima_list = jax.vmap(get_optima)(jnp.arange(num_optima))

    def get_c_val(i):
        alpha = 1000.0 ** (2.0 * (i - 1) / (num_optima - 2))
        c_mat = lambda_alpha_trafo(alpha, dim) / (alpha ** 0.25)
        return c_mat * (i != 0) + lambda_alpha_trafo(1000, dim) * (i == 0)

    c_list = jax.vmap(get_c_val)(jnp.arange(num_optima))
    max_value = jnp.array([-1.0])

    for i in range(num_optima):
        w = 10 * (i == 0) + (1.1 + 8.0 * (i - 1.0) / (num_optima - 2.0)) * (
            i != 0
        )
        diff = jnp.matmul(R, (arr - optima_list[i]).reshape(-1, 1))
        e = jnp.matmul(diff.transpose(), jnp.matmul(c_list[i], diff))
        max_value = jnp.maximum(max_value, w * jnp.exp(-e / (2.0 * dim)))
    return oscillation_trafo(
        10.0 - max_value.squeeze()
    ) ** 2 + boundary_penalty(arr, dim)


def Katsuura(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Katsuura fn - multi-modal weak (p.115; Hansen et al., 2009)."""
    # NOTE: Numerically instable - need higher precision float64
    # from jax.config import config
    # config.update("jax_enable_x64", True)
    dim = arr.shape[0]
    r_x = jnp.matmul(R, arr)
    z_vec = jnp.matmul(lambda_alpha_trafo(100.0, dim), r_x)
    z_vec = jnp.matmul(Q, z_vec)

    def get_val(i):
        s = 0.0
        for j in range(1, 33):
            s += (
                jnp.abs(2 ** j * z_vec[i] - jnp.round(2 ** j * z_vec[i]))
                / 2 ** j
            )
        return (1 + (i + 1) * s) ** (10.0 / dim ** 1.2)

    out = jax.vmap(get_val)(jnp.arange(dim))
    mask = jnp.arange(dim) < dim
    coeff = 1.0 - mask  # 1st to zero and add 1 for multiply
    prod = jnp.prod(out * mask + coeff)
    return (
        (10.0 / dim ** 2) * prod - 10.0 / dim ** 2 + boundary_penalty(arr, dim)
    )


def Lunacek(arr: chex.Array, R: chex.Array, Q: chex.Array) -> chex.Array:
    """Lunacek fn - multi-modal weak (p.120; Hansen et al., 2009)."""
    dim = arr.shape[0]
    mu0 = 2.5
    s = 1.0 - 1.0 / (2.0 * (dim + 20.0) ** 0.5 - 8.2)
    mu1 = -(((mu0 ** 2 - 1) / s) ** 0.5)
    x_opt = jnp.array([mu0 / 2] * dim)
    x_hat = 2 * arr * jnp.sign(x_opt)
    x_vec = x_hat - mu0
    x_vec = jnp.matmul(R, x_vec)
    z_vec = jnp.matmul(lambda_alpha_trafo(100, dim), x_vec)
    z_vec = jnp.matmul(Q, z_vec)

    s1 = jnp.sum((x_hat - mu0) ** 2)
    s2 = jnp.sum((x_hat - mu1) ** 2)
    s3 = jnp.sum(jnp.cos(2 * jnp.pi * z_vec))
    return (
        jnp.minimum(s1, dim + s * s2)
        + 10.0 * (dim - s3)
        + 10 ** 4 * boundary_penalty(arr, dim)
    )


BBOB_fns = {
    # Part 1: Separable functions
    "Sphere": Sphere,
    "EllipsoidalOriginal": EllipsoidalOriginal,
    "RastriginOriginal": RastriginOriginal,
    "BuecheRastrigin": BuecheRastrigin,
    "LinearSlope": LinearSlope,
    # Part 2: Functions with low or moderate conditions
    "AttractiveSector": AttractiveSector,
    "StepEllipsoidal": StepEllipsoidal,
    "RosenbrockOriginal": RosenbrockOriginal,
    "RosenbrockRotated": RosenbrockRotated,
    # Part 3: Functions with high conditioning and unimodal
    "EllipsoidalRotated": EllipsoidalRotated,
    "Discus": Discus,
    "BentCigar": BentCigar,
    "SharpRidge": SharpRidge,
    "DifferentPowers": DifferentPowers,
    # Part 4: Multi-modal functions with adequate global structure
    "RastriginRotated": RastriginRotated,
    "Weierstrass": Weierstrass,
    "SchaffersF7": SchaffersF7,
    "SchaffersF7IllConditioned": SchaffersF7IllConditioned,
    "GriewankRosenbrock": GriewankRosenbrock,
    # Part 5: Multi-modal functions with weak global structure
    "Schwefel": Schwefel,
    "Katsuura": Katsuura,
    "Lunacek": Lunacek,
    "Gallagher101Me": Gallagher101Me,
    "Gallagher21Hi": Gallagher21Hi,
}



from flax import linen as nn
import jax
import chex
from typing import Tuple, Optional
from .shared import (
    identity_out,
    tanh_out,
    categorical_out,
    gaussian_out,
    default_bias_init,
    kernel_init_fn,
)


class LSTM(nn.Module):
    """Simple LSTM Wrapper with flexible output head."""

    num_hidden_units: int = 32
    num_output_units: int = 1
    output_activation: str = "identity"
    kernel_init_type: str = "lecun_normal"
    model_name: str = "LSTM"

    @nn.compact
    def __call__(
        self,
        x: chex.Array,
        carry: chex.ArrayTree,
        rng: Optional[chex.PRNGKey] = None,
    ) -> Tuple[Tuple[chex.ArrayTree, chex.ArrayTree], chex.Array]:
        lstm_state, x = nn.LSTMCell(
            bias_init=default_bias_init(),
            kernel_init=kernel_init_fn[self.kernel_init_type](),
        )(carry, x)
        if self.output_activation == "identity":
            x = identity_out(x, self.num_output_units, self.kernel_init_type)
        elif self.output_activation == "tanh":
            x = tanh_out(x, self.num_output_units, self.kernel_init_type)
        # Categorical and gaussian output heads require rng for sampling
        elif self.output_activation == "categorical":
            x = categorical_out(
                rng, x, self.num_output_units, self.kernel_init_type
            )
        elif self.output_activation == "gaussian":
            x = gaussian_out(
                rng, x, self.num_output_units, self.kernel_init_type
            )
        return lstm_state, x

    def initialize_carry(self) -> Tuple[chex.ArrayTree, chex.ArrayTree]:
        """Initialize hidden state of LSTM."""
        return nn.LSTMCell.initialize_carry(
            jax.random.PRNGKey(0), (), self.num_hidden_units
        )



import jax
import jax.numpy as jnp
import chex
from typing import Tuple


def get_best_fitness_member(
    x: chex.Array, fitness: chex.Array, state, maximize: bool = False
) -> Tuple[chex.Array, float]:
    """Check if fitness improved & replace in ES state."""
    fitness_min = jax.lax.select(maximize, -1 * fitness, fitness)
    max_and_later = maximize and state.gen_counter > 0
    best_fit_min = jax.lax.select(
        max_and_later, -1 * state.best_fitness, state.best_fitness
    )
    best_in_gen = jnp.argmin(fitness_min)
    best_in_gen_fitness, best_in_gen_member = (
        fitness_min[best_in_gen],
        x[best_in_gen],
    )
    replace_best = best_in_gen_fitness < best_fit_min
    best_fitness = jax.lax.select(
        replace_best, best_in_gen_fitness, best_fit_min
    )
    best_member = jax.lax.select(
        replace_best, best_in_gen_member, state.best_member
    )
    best_fitness = jax.lax.select(maximize, -1 * best_fitness, best_fitness)
    return best_member, best_fitness



import functools
from typing import List
import chex
from flax import struct
import jax
import jax.numpy as jnp
from evosax.strategies.snes import get_snes_weights
from evosax.core.fitness import centered_rank_trafo


class TraceConstructor(object):
    def __init__(self, num_dims: int, timescales: chex.Array):
        self.num_dims = num_dims
        self.timescales = timescales

    def initialize(self) -> chex.Array:
        return jnp.zeros((self.num_dims, self.timescales.shape[0]))

    def update(self, paths: chex.Array, diff: chex.Array) -> chex.Array:
        def update_path(lrate, path, diff):
            return (1 - lrate) * path + lrate * diff

        return jax.vmap(update_path, in_axes=(0, 1, None), out_axes=1)(
            self.timescales, paths, diff
        )


@struct.dataclass
class DistributionFeaturesState:
    old_mean: chex.Array
    old_sigma: chex.Array
    momentum_mean: chex.Array
    momentum_sigma: chex.Array
    evopath_mean: chex.Array
    evopath_sigma: chex.Array


class DistributionFeaturizer(object):
    def __init__(
        self,
        popsize: int,
        num_dims: int,
        seq_len: int,
        use_mean: bool = False,
        use_sigma: bool = False,
        use_evopaths: bool = False,
        use_momentum: bool = False,
        evopath_timescales: List[float] = [0.9, 0.95, 1.0],
        momentum_timescales: List[float] = [0.9, 0.95, 1.0],
        use_oai_grad: bool = False,
        verbose: bool = False,
    ):
        self.popsize = popsize
        self.num_dims = num_dims
        self.seq_len = seq_len
        self.use_mean = use_mean
        self.use_sigma = use_sigma
        self.use_evopaths = use_evopaths
        self.use_momentum = use_momentum
        self.evopath_timescales = evopath_timescales
        self.momentum_timescales = momentum_timescales
        self.use_oai_grad = use_oai_grad
        self.verbose = verbose

        self.evopath = TraceConstructor(num_dims, jnp.array(evopath_timescales))
        self.mompath = TraceConstructor(num_dims, jnp.array(momentum_timescales))

        if self.verbose:
            print(
                f"Distribution Features / Batch shape: {self.num_features} / {self.example_batch_shape}"
            )
            print("[BASE] SNES grad mean -> sum w_i (x-mu)/sigma")
            print("[BASE] SNES grad sigma -> sum w_i [(x-mu)/sigma]**2 - 1")
            print(f"[{self.use_mean}] Mean")
            print(f"[{self.use_sigma}] Sigma")
            print(f"[{self.use_evopaths}] Evolution Paths -> {self.evopath_timescales}")
            print(f"[{self.use_momentum}] Momentum -> {self.momentum_timescales}")
            print(f"[{self.use_oai_grad}] OpenAI Gradient")

    @functools.partial(jax.jit, static_argnums=0)
    def initialize(self) -> DistributionFeaturesState:
        return DistributionFeaturesState(
            old_mean=jnp.zeros((self.num_dims,)),
            old_sigma=jnp.ones((self.num_dims,)),
            momentum_mean=self.mompath.initialize(),
            momentum_sigma=self.mompath.initialize(),
            evopath_mean=self.evopath.initialize(),
            evopath_sigma=self.evopath.initialize(),
        )

    @functools.partial(jax.jit, static_argnums=0)
    def featurize(
        self,
        x: chex.Array,
        fitness: chex.Array,
        mean: chex.Array,
        sigma: chex.Array,
        state: DistributionFeaturesState,
    ) -> chex.Array:
        ranks = fitness.argsort()
        weights = get_snes_weights(fitness.shape[0])
        noise = (x - mean) / sigma
        sorted_noise = noise[ranks]
        grad_mean = (weights * sorted_noise).sum(axis=0).reshape(-1, 1)
        grad_sigma = (weights * (sorted_noise**2 - 1)).sum(axis=0).reshape(-1, 1)
        distrib_features = jnp.concatenate([grad_mean, grad_sigma], axis=1)
        if self.use_mean:
            distrib_features = jnp.concatenate(
                [distrib_features, mean.reshape(-1, 1)], axis=1
            )
        if self.use_sigma:
            distrib_features = jnp.concatenate(
                [distrib_features, sigma.reshape(-1, 1)], axis=1
            )

        # Update traces
        path_mean = self.evopath.update(state.evopath_mean, grad_mean.squeeze())
        path_sigma = self.evopath.update(state.evopath_sigma, grad_sigma.squeeze())
        if self.use_evopaths:
            distrib_features = jnp.concatenate(
                [distrib_features, path_mean, path_sigma], axis=1
            )
        mom_mean = self.mompath.update(state.momentum_mean, mean - state.old_mean)
        mom_sigma = self.mompath.update(state.momentum_sigma, sigma - state.old_sigma)
        if self.use_momentum:
            distrib_features = jnp.concatenate(
                [distrib_features, mom_mean, mom_sigma], axis=1
            )
        if self.use_oai_grad:
            popsize = x.shape[0]
            fitness_re = centered_rank_trafo(fitness)
            oai_grad = 1.0 / (popsize * sigma) * jnp.dot(noise.T, fitness_re)
            distrib_features = jnp.concatenate(
                [distrib_features, oai_grad.reshape(-1, 1)], axis=1
            )
        return distrib_features, DistributionFeaturesState(
            mean, sigma, mom_mean, mom_sigma, path_mean, path_sigma
        )

    @property
    def num_features(self):
        return (
            2  # SNES grad mean, sigma
            + 2 * len(self.evopath_timescales) * self.use_evopaths  # evopaths
            + 2 * len(self.momentum_timescales) * self.use_momentum  # momentum
            + self.use_mean  # mean
            + self.use_sigma  # sigma
            + self.use_oai_grad  # oai grad
        )

    @property
    def example_batch_shape(self):
        return (
            1,  # Batchsize
            self.seq_len,  # Timesteps
            self.num_dims,  # Number of dims
            self.num_features,  # Distribution features
        )



import jax
import jax.numpy as jnp
from evosax.networks import MLP, LSTM, CNN, All_CNN_C


def test_mlp_forward():
    rng = jax.random.PRNGKey(0)
    network = MLP(
        num_hidden_units=64,
        num_hidden_layers=2,
        num_output_units=1,
        hidden_activation="relu",
        output_activation="identity",
    )
    pholder = jnp.zeros((4,))
    params = network.init(
        rng,
        x=pholder,
        rng=rng,
    )
    out = jax.jit(network.apply)(params, pholder, rng)
    assert out.shape == (1,)


# def test_lstm_forward():
#     rng = jax.random.PRNGKey(0)
#     network = LSTM(
#         num_hidden_units=32,
#         num_output_units=1,
#         output_activation="identity",
#     )
#     pholder = jnp.zeros((4,))
#     carry_init = network.initialize_carry()
#     params = network.init(
#         rng,
#         x=pholder,
#         carry=carry_init,
#         rng=rng,
#     )
#     _, out = jax.jit(network.apply)(params, pholder, carry_init, rng)
#     assert out.shape == (1,)


def test_cnn_forward():
    rng = jax.random.PRNGKey(0)
    network = CNN(
        depth_1=1,
        depth_2=1,
        features_1=16,
        features_2=8,
        kernel_1=3,
        kernel_2=5,
        strides_1=1,
        strides_2=1,
        num_linear_layers=1,
        num_hidden_units=16,
        num_output_units=10,
    )
    pholder = jnp.zeros((1, 28, 28, 1))
    params = network.init(
        rng,
        x=pholder,
        rng=rng,
    )
    out = jax.jit(network.apply)(params, pholder, rng)
    assert out.shape == (1, 10)


def test_all_cnn_forward():
    rng = jax.random.PRNGKey(0)
    network = All_CNN_C(
        depth_1=1,
        depth_2=1,
        features_1=16,
        features_2=8,
        kernel_1=3,
        kernel_2=5,
        strides_1=1,
        strides_2=1,
        final_window=(28, 28),
        num_output_units=10,
    )
    pholder = jnp.zeros((1, 28, 28, 1))
    params = network.init(
        rng,
        x=pholder,
        rng=rng,
    )
    out = jax.jit(network.apply)(params, pholder, rng)
    assert out.shape == (1, 10)



import jax
import jax.numpy as jnp
from typing import Optional
import chex


class GymnaxFitness(object):
    def __init__(
        self,
        env_name: str = "CartPole-v1",
        num_env_steps: Optional[int] = None,
        num_rollouts: int = 16,
        env_kwargs: dict = {},
        env_params: dict = {},
        test: bool = False,
        n_devices: Optional[int] = None,
    ):
        self.env_name = env_name
        self.num_rollouts = num_rollouts
        self.test = test

        try:
            import gymnax
        except ImportError:
            raise ImportError(
                "You need to install `gymnax` to use its fitness rollouts."
            )

        # Define the RL environment & replace default parameters if desired
        self.env, self.env_params = gymnax.make(env_name, **env_kwargs)
        self.env_params.replace(**env_params)

        if num_env_steps is None:
            self.num_env_steps = self.env_params.max_steps_in_episode
        else:
            self.num_env_steps = num_env_steps
        self.steps_per_member = self.num_env_steps * num_rollouts

        self.action_shape = self.env.num_actions
        self.input_shape = self.env.observation_space(self.env_params).shape
        if n_devices is None:
            self.n_devices = jax.local_device_count()
        else:
            self.n_devices = n_devices

        # Keep track of total steps executed in environment
        self.total_env_steps = 0

    def set_apply_fn(self, network_apply, carry_init=None):
        """Set the network forward function."""
        self.network = network_apply
        # Set rollout function based on model architecture
        if carry_init is not None:
            self.single_rollout = self.rollout_rnn
            self.carry_init = carry_init
        else:
            self.single_rollout = self.rollout_ffw
        self.rollout_repeats = jax.vmap(self.single_rollout, in_axes=(0, None))
        self.rollout_pop = jax.vmap(self.rollout_repeats, in_axes=(None, 0))
        # pmap over popmembers if > 1 device is available - otherwise pmap
        if self.n_devices > 1:
            self.rollout_map = self.rollout_pmap
            print(
                f"GymFitness: {self.n_devices} devices detected. Please make"
                " sure that the ES population size divides evenly across the"
                " number of devices to pmap/parallelize over."
            )
        else:
            self.rollout_map = self.rollout_pop

    def rollout_pmap(
        self, rng_input: chex.PRNGKey, policy_params: chex.ArrayTree
    ):
        """Parallelize rollout across devices. Split keys/reshape correctly."""
        keys_pmap = jnp.tile(rng_input, (self.n_devices, 1, 1))
        rew_dev, steps_dev = jax.pmap(self.rollout_pop)(
            keys_pmap, policy_params
        )
        rew_re = rew_dev.reshape(-1, self.num_rollouts)
        steps_re = steps_dev.reshape(-1, self.num_rollouts)
        return rew_re, steps_re

    def rollout(self, rng_input: chex.PRNGKey, policy_params: chex.ArrayTree):
        """Placeholder fn call for rolling out a population for multi-evals."""
        rng_pop = jax.random.split(rng_input, self.num_rollouts)
        scores, masks = jax.jit(self.rollout_map)(rng_pop, policy_params)
        # Update total step counter using only transitions before termination
        self.total_env_steps += masks.sum()
        return scores

    def rollout_ffw(
        self, rng_input: chex.PRNGKey, policy_params: chex.ArrayTree
    ):
        """Rollout an episode with lax.scan."""
        # Reset the environment
        rng_reset, rng_episode = jax.random.split(rng_input)
        obs, state = self.env.reset(rng_reset, self.env_params)

        def policy_step(state_input, tmp):
            """lax.scan compatible step transition in jax env."""
            obs, state, policy_params, rng, cum_reward, valid_mask = state_input
            rng, rng_step, rng_net = jax.random.split(rng, 3)
            action = self.network(policy_params, obs, rng=rng_net)
            next_o, next_s, reward, done, _ = self.env.step(
                rng_step, state, action, self.env_params
            )
            new_cum_reward = cum_reward + reward * valid_mask
            new_valid_mask = valid_mask * (1 - done)
            carry = [
                next_o.squeeze(),
                next_s,
                policy_params,
                rng,
                new_cum_reward,
                new_valid_mask,
            ]
            y = [new_valid_mask]
            return carry, y

        # Scan over episode step loop
        carry_out, scan_out = jax.lax.scan(
            policy_step,
            [
                obs,
                state,
                policy_params,
                rng_episode,
                jnp.array([0.0]),
                jnp.array([1.0]),
            ],
            (),
            self.num_env_steps,
        )
        # Return the sum of rewards accumulated by agent in episode rollout
        ep_mask = scan_out
        cum_return = carry_out[-2].squeeze()
        return cum_return, jnp.array(ep_mask)

    def rollout_rnn(
        self, rng_input: chex.PRNGKey, policy_params: chex.ArrayTree
    ):
        """Rollout a jitted episode with lax.scan."""
        # Reset the environment
        rng, rng_reset = jax.random.split(rng_input)
        obs, state = self.env.reset(rng_reset, self.env_params)
        hidden = self.carry_init()

        def policy_step(state_input, tmp):
            """lax.scan compatible step transition in jax env."""
            (
                obs,
                state,
                policy_params,
                rng,
                hidden,
                cum_reward,
                valid_mask,
            ) = state_input
            rng, rng_step, rng_net = jax.random.split(rng, 3)
            hidden, action = self.network(policy_params, obs, hidden, rng_net)
            next_o, next_s, reward, done, _ = self.env.step(
                rng_step, state, action, self.env_params
            )
            new_cum_reward = cum_reward + reward * valid_mask
            new_valid_mask = valid_mask * (1 - done)
            carry, y = [
                next_o.squeeze(),
                next_s,
                policy_params,
                rng,
                hidden,
                new_cum_reward,
                new_valid_mask,
            ], [new_valid_mask]
            return carry, y

        # Scan over episode step loop
        carry_out, scan_out = jax.lax.scan(
            policy_step,
            [
                obs,
                state,
                policy_params,
                rng,
                hidden,
                jnp.array([0.0]),
                jnp.array([1.0]),
            ],
            (),
            self.num_env_steps,
        )
        # Return masked sum of rewards accumulated by agent in episode
        ep_mask = scan_out
        cum_return = carry_out[-2].squeeze()
        return cum_return, jnp.array(ep_mask)



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from ..utils import get_best_fitness_member


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    sigma_init: float = 1.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


class HillClimber(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Simple Gaussian Hill Climbing"""
        super().__init__(
            popsize, num_dims, pholder_params, mean_decay, n_devices, **fitness_kwargs
        )
        self.strategy_name = "HillClimber"

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        return EvoParams()

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """`initialize` the evolution strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init * jnp.ones((self.num_dims,)),
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new proposed candidates to evaluate next."""
        # Sampling of N(0, 1) noise
        z = jax.random.normal(
            rng,
            (self.popsize, self.num_dims),
        )
        x = state.best_member + state.sigma.reshape(1, self.num_dims) * z
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` update to ES state."""
        best_member, best_fitness = get_best_fitness_member(x, fitness, state, False)
        return state.replace(mean=best_member)



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from ..utils.eigen_decomp import diag_eigen_decomp


@struct.dataclass
class EvoState:
    p_sigma: chex.Array
    p_c: chex.Array
    C: chex.Array
    D: Optional[chex.Array]
    mean: chex.Array
    sigma: chex.Array
    sigma_scale: float
    weights: chex.Array
    weights_truncated: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    mu_eff: float
    c_1: float
    c_mu: float
    c_sigma: float
    d_sigma: float
    c_c: float
    chi_n: float
    c_m: float = 1.0
    sigma_init: float = 0.065
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


def get_cma_elite_weights(
    popsize: int, elite_popsize: int
) -> Tuple[chex.Array, chex.Array]:
    """Utility helper to create truncated elite weights for mean
    update and full weights for covariance update."""
    weights_prime = jnp.array(
        [jnp.log(elite_popsize + 1) - jnp.log(i + 1) for i in range(elite_popsize)]
    )
    weights = weights_prime / jnp.sum(weights_prime)
    weights_truncated = jnp.zeros(popsize)
    weights_truncated = weights_truncated.at[:elite_popsize].set(weights)
    return weights, weights_truncated


class Sep_CMA_ES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        elite_ratio: float = 0.5,
        sigma_init: float = 1.0,
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Separable CMA-ES (e.g. Ros & Hansen, 2008)
        Reference: https://hal.inria.fr/inria-00287367/document
        Inspired by: github.com/CyberAgentAILab/cmaes/blob/main/cmaes/_sepcma.py
        """
        super().__init__(
            popsize,
            num_dims,
            pholder_params,
            mean_decay,
            n_devices,
            **fitness_kwargs,
        )
        assert 0 <= elite_ratio <= 1
        self.elite_ratio = elite_ratio
        self.elite_popsize = max(1, int(self.popsize * self.elite_ratio))
        self.strategy_name = "Sep_CMA_ES"

        # Set core kwargs es_params
        self.sigma_init = sigma_init

        # Robustness for int32 - squaring in hyperparameter calculations
        self.max_dims_sq = jnp.minimum(self.num_dims, 40000)

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolution strategy."""
        # Temporarily create elite weights for rest of parameters
        weights, _ = get_cma_elite_weights(self.popsize, self.elite_popsize)
        mu_eff = 1 / jnp.sum(weights**2)

        # lrates for rank-one and rank-μ C updates
        alpha_cov = 2
        c_1 = alpha_cov / ((self.max_dims_sq + 1.3) ** 2 + mu_eff)
        c_mu_full = 2 / mu_eff / ((self.max_dims_sq + jnp.sqrt(2)) ** 2) + (
            1 - 1 / mu_eff
        ) * jnp.minimum(1, (2 * mu_eff - 1) / ((self.max_dims_sq + 2) ** 2 + mu_eff))
        c_mu = (self.num_dims + 2) / 3 * c_mu_full

        # lrate for cumulation of step-size control and rank-one update
        c_sigma = (mu_eff + 2) / (self.num_dims + mu_eff + 3)
        d_sigma = (
            1
            + 2 * jnp.maximum(0, jnp.sqrt((mu_eff - 1) / (self.num_dims + 1)) - 1)
            + c_sigma
        )
        c_c = 4 / (self.num_dims + 4)
        chi_n = jnp.sqrt(self.num_dims) * (
            1.0 - (1.0 / (4.0 * self.num_dims)) + 1.0 / (21.0 * (self.max_dims_sq**2))
        )
        params = EvoParams(
            mu_eff=mu_eff,
            c_1=c_1,
            c_mu=c_mu,
            c_sigma=c_sigma,
            d_sigma=d_sigma,
            c_c=c_c,
            chi_n=chi_n,
            sigma_init=self.sigma_init,
        )
        return params

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """`initialize` the evolution strategy."""
        # Population weightings - store in state
        weights, weights_truncated = get_cma_elite_weights(
            self.popsize, self.elite_popsize
        )
        # Initialize evolution paths & covariance matrix
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        state = EvoState(
            p_sigma=jnp.zeros(self.num_dims),
            p_c=jnp.zeros(self.num_dims),
            sigma_scale=params.sigma_init,
            sigma=jnp.ones(self.num_dims) * params.sigma_init,
            mean=initialization,
            C=jnp.ones(self.num_dims),
            D=diag_eigen_decomp(jnp.ones(self.num_dims), None),
            weights=weights,
            weights_truncated=weights_truncated,
            best_member=initialization,
        )
        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        x = sample(
            rng,
            state.mean,
            state.sigma_scale,
            state.D,
            self.num_dims,
            self.popsize,
        )
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        # Sort new results, extract elite, store best performer
        concat_p_f = jnp.hstack([jnp.expand_dims(fitness, 1), x])
        sorted_solutions = concat_p_f[concat_p_f[:, 0].argsort()]
        # Update mean, isotropic/anisotropic paths, covariance, stepsize
        y_k, y_w, mean = update_mean(
            state.mean,
            state.sigma,
            sorted_solutions,
            params.c_m,
            state.weights_truncated,
        )

        p_sigma, D = update_p_sigma(
            state.C, state.D, state.p_sigma, y_w, params.c_sigma, params.mu_eff
        )

        p_c, norm_p_sigma, h_sigma = update_p_c(
            mean,
            p_sigma,
            state.p_c,
            state.gen_counter + 1,
            y_w,
            params.c_sigma,
            params.c_c,
            params.chi_n,
            params.mu_eff,
        )

        C = update_covariance(
            p_c,
            state.C,
            y_k,
            h_sigma,
            state.weights,
            params.c_c,
            params.c_1,
            params.c_mu,
        )
        sigma_scale = update_sigma(
            state.sigma_scale,
            norm_p_sigma,
            params.c_sigma,
            params.d_sigma,
            params.chi_n,
        )
        sigma = sigma_scale * D
        return state.replace(
            mean=mean,
            p_sigma=p_sigma,
            C=C,
            D=D,
            p_c=p_c,
            sigma_scale=sigma_scale,
            sigma=sigma,
        )


def update_mean(
    mean: chex.Array,
    sigma: float,
    sorted_solutions: chex.Array,
    c_m: float,
    weights_truncated: chex.Array,
) -> Tuple[chex.Array, chex.Array, chex.Array]:
    """Update mean of strategy."""
    x_k = sorted_solutions[:, 1:]  # ~ N(m, σ^2 C)
    y_k = (x_k - mean) / sigma  # ~ N(0, C)
    y_w = jnp.sum(y_k.T * weights_truncated, axis=1)
    mean += c_m * sigma * y_w
    return y_k, y_w, mean


def update_p_sigma(
    C: chex.Array,
    D: chex.Array,
    p_sigma: chex.Array,
    y_w: chex.Array,
    c_sigma: float,
    mu_eff: float,
) -> Tuple[chex.Array, None]:
    """Update evolution path for covariance matrix."""
    D = diag_eigen_decomp(C, D)
    p_sigma_new = (1 - c_sigma) * p_sigma + jnp.sqrt(
        c_sigma * (2 - c_sigma) * mu_eff
    ) * (y_w / D)
    return p_sigma_new, D


def update_p_c(
    mean: chex.Array,
    p_sigma: chex.Array,
    p_c: chex.Array,
    gen_counter: int,
    y_w: chex.Array,
    c_sigma: float,
    c_c: float,
    chi_n: float,
    mu_eff: float,
) -> Tuple[chex.Array, float, float]:
    """Update evolution path for sigma/stepsize."""
    norm_p_sigma = jnp.linalg.norm(p_sigma)
    h_sigma_cond_left = norm_p_sigma / jnp.sqrt(
        1 - (1 - c_sigma) ** (2 * (gen_counter))
    )
    h_sigma_cond_right = (1.4 + 2 / (mean.shape[0] + 1)) * chi_n
    h_sigma = 1.0 * (h_sigma_cond_left < h_sigma_cond_right)
    p_c_new = (1 - c_c) * p_c + h_sigma * jnp.sqrt(c_c * (2 - c_c) * mu_eff) * y_w
    return p_c_new, norm_p_sigma, h_sigma


def update_covariance(
    p_c: chex.Array,
    C: chex.Array,
    y_k: chex.Array,
    h_sigma: float,
    weights: chex.Array,
    c_c: float,
    c_1: float,
    c_mu: float,
) -> chex.Array:
    """Update cov. matrix estimator using rank 1 + μ updates."""
    delta_h_sigma = (1 - h_sigma) * c_c * (2 - c_c)
    rank_one = p_c**2
    rank_mu = jnp.sum(
        jnp.array([w * (y**2) for w, y in zip(weights, y_k)]),
        axis=0,
    )
    C = (
        (1 + c_1 * delta_h_sigma - c_1 - c_mu * jnp.sum(weights)) * C
        + c_1 * rank_one
        + c_mu * rank_mu
    )
    return C


def update_sigma(
    sigma: float,
    norm_p_sigma: float,
    c_sigma: float,
    d_sigma: float,
    chi_n: float,
) -> float:
    """Update stepsize sigma."""
    sigma_new = sigma * jnp.exp((c_sigma / d_sigma) * (norm_p_sigma / chi_n - 1))
    return sigma_new


def sample(
    rng: chex.PRNGKey,
    mean: chex.Array,
    sigma_scale: float,
    D: chex.Array,
    n_dim: int,
    pop_size: int,
) -> chex.Array:
    """Jittable Gaussian Sample Helper."""
    z = jax.random.normal(rng, (n_dim, pop_size))  # ~ N(0, I)
    y = jnp.diag(D).dot(z)  # ~ N(0, C)
    y = jnp.swapaxes(y, 1, 0)
    x = mean + sigma_scale * y  # ~ N(m, σ^2 C)
    return x



from typing import Tuple, Optional, Union
import jax
import jax.numpy as jnp
import chex
from flax import struct
from ..strategy import Strategy
from .des import get_des_weights


@struct.dataclass
class EvoState:
    mean: chex.Array
    sigma: chex.Array
    weights: chex.Array
    best_member: chex.Array
    best_fitness: float = jnp.finfo(jnp.float32).max
    gen_counter: int = 0


@struct.dataclass
class EvoParams:
    lrate_mean: float = 1.0
    lrate_sigma: float = 1.0
    sigma_init: float = 1.0
    temperature: float = 0.0
    init_min: float = 0.0
    init_max: float = 0.0
    clip_min: float = -jnp.finfo(jnp.float32).max
    clip_max: float = jnp.finfo(jnp.float32).max


def get_snes_weights(popsize: int, use_baseline: bool = True):
    """Get recombination weights for different ranks."""

    def get_weight(i):
        return jnp.maximum(0, jnp.log(popsize / 2 + 1) - jnp.log(i))

    weights = jax.vmap(get_weight)(jnp.arange(1, popsize + 1))
    weights_norm = weights / jnp.sum(weights)
    return (weights_norm - use_baseline * (1 / popsize))[:, None]


class SNES(Strategy):
    def __init__(
        self,
        popsize: int,
        num_dims: Optional[int] = None,
        pholder_params: Optional[Union[chex.ArrayTree, chex.Array]] = None,
        sigma_init: float = 1.0,
        temperature: float = 0.0,  # good values tend to be between 12 and 20
        mean_decay: float = 0.0,
        n_devices: Optional[int] = None,
        **fitness_kwargs: Union[bool, int, float]
    ):
        """Separable Exponential Natural ES (Wierstra et al., 2014)
        Reference: https://www.jmlr.org/papers/volume15/wierstra14a/wierstra14a.pdf
        """
        super().__init__(
            popsize, num_dims, pholder_params, mean_decay, n_devices, **fitness_kwargs
        )
        self.strategy_name = "SNES"

        # Set core kwargs es_params
        self.sigma_init = sigma_init
        self.temperature = temperature

    @property
    def params_strategy(self) -> EvoParams:
        """Return default parameters of evolutionary strategy."""
        lrate_sigma = (3 + jnp.log(self.num_dims)) / (5 * jnp.sqrt(self.num_dims))
        params = EvoParams(
            lrate_sigma=lrate_sigma,
            sigma_init=self.sigma_init,
            temperature=self.temperature,
        )
        return params

    def initialize_strategy(self, rng: chex.PRNGKey, params: EvoParams) -> EvoState:
        """`initialize` the evolutionary strategy."""
        initialization = jax.random.uniform(
            rng,
            (self.num_dims,),
            minval=params.init_min,
            maxval=params.init_max,
        )
        use_des_weights = params.temperature > 0.0
        weights = jax.lax.select(
            use_des_weights,
            get_des_weights(self.popsize, params.temperature),
            get_snes_weights(self.popsize),
        )
        state = EvoState(
            mean=initialization,
            sigma=params.sigma_init * jnp.ones(self.num_dims),
            weights=weights,
            best_member=initialization,
        )

        return state

    def ask_strategy(
        self, rng: chex.PRNGKey, state: EvoState, params: EvoParams
    ) -> Tuple[chex.Array, EvoState]:
        """`ask` for new parameter candidates to evaluate next."""
        noise = jax.random.normal(rng, (self.popsize, self.num_dims))
        x = state.mean + noise * state.sigma.reshape(1, self.num_dims)
        return x, state

    def tell_strategy(
        self,
        x: chex.Array,
        fitness: chex.Array,
        state: EvoState,
        params: EvoParams,
    ) -> EvoState:
        """`tell` performance data for strategy state update."""
        s = (x - state.mean) / state.sigma
        ranks = fitness.argsort()
        sorted_noise = s[ranks]
        grad_mean = (state.weights * sorted_noise).sum(axis=0)
        grad_sigma = (state.weights * (sorted_noise**2 - 1)).sum(axis=0)
        mean = state.mean + params.lrate_mean * state.sigma * grad_mean
        sigma = state.sigma * jnp.exp(params.lrate_sigma / 2 * grad_sigma)
        return state.replace(mean=mean, sigma=sigma)
